"""A data table gadget."""

from __future__ import annotations

from collections.abc import Callable, Iterator, Sequence
from dataclasses import dataclass, replace
from enum import Enum
from itertools import count, islice
from typing import Any, Literal, Protocol

from uwcwidth import wcswidth

from ..terminal.events import MouseEvent
from ..text_tools import egc_ord
from .behaviors.button_behavior import ButtonBehavior
from .behaviors.themable import Themable
from .gadget import (
    Gadget,
    Point,
    Pointlike,
    PosHint,
    Size,
    SizeHint,
    Sizelike,
    _GadgetList,
)
from .grid_layout import GridLayout
from .pane import Pane
from .scroll_view import ScrollView
from .text import Text, add_text

__all__ = ["ColumnStyle", "DataTable", "Point", "Size"]


class SupportsLessThan(Protocol):
    """Supports the less than (`<`) operator."""

    def __lt__(self, other: Any, /) -> bool: ...


class _ColumnLabels(GridLayout):
    children: _GadgetList[_ColumnLabel]


class _DataCells(GridLayout):
    children: _GadgetList[_DataCell]


@dataclass
class ColumnStyle:
    """
    Style of a column in a data table. Includes how data is rendered,
    alignment, padding, and minimum width.

    Parameters
    ----------
    render : Callable[[T], str] | None, default: None
        A callable that renders column data into a string. Uses the built-in ``str`` by
        default.
    alignment : Literal["center", "left", "right"], default: "left"
        Alignment of the column.
    padding : int, default: 1
        Left and right padding of column.
    min_width : int, default: 0
        Minimum width of column.
    allow_sorting : bool, default: True
        Whether sorting is allowed for column.

    Attributes
    ----------
    render : Callable[[T], str]
        A callable that renders column data into a string.
    alignment : Literal["center", "left", "right"]
        Alignment of the column.
    padding : int
        Left and right padding of column.
    min_width : int
        Minimum width of column.
    allow_sorting : bool
        Whether sorting is allowed for column.
    """

    render: Callable[[SupportsLessThan], str]
    """
    A callable that renders column data into a string. Uses the built-in `str` by
    default.
    """
    alignment: Literal["center", "left", "right"]
    """Alignment of the column."""
    padding: int
    """Left and right padding of column."""
    min_width: int
    """Minimum width of column."""
    allow_sorting: bool
    """Whether sorting is allowed for column."""

    def __init__(
        self,
        render: Callable[[SupportsLessThan], str] | None = None,
        alignment: Literal["center", "left", "right"] = "left",
        padding: int = 1,
        min_width: int = 0,
        allow_sorting: bool = True,
    ):
        self.render = str if render is None else render
        self.alignment = alignment
        self.padding = padding
        self.min_width = min_width
        self.allow_sorting = allow_sorting


class _SortState(str, Enum):
    """Sorted state of a column in a data table."""

    NOT_SORTED = "↕"
    ASCENDING = "↑"
    DESCENDING = "↓"


_SORT_INDICATOR_SPACING = 1
"""Spaces between column label text and sort indicator."""
_SORT_INDICATOR_WIDTH = wcswidth(_SortState.NOT_SORTED.value)
"""Character width of sort indicator. (Indicator values should be same width.)"""
_ALIGN_FORMATTER = {"center": "^", "left": "<", "right": ">"}
"""Convert an alignment to f-string format specification."""


class _CellBase(ButtonBehavior, Text):
    """Base for cells in a data table."""

    def __init__(self, data_table: DataTable, column_id: int, **kwargs):
        super().__init__(**kwargs)
        self.data_table = data_table
        self.column_id = column_id
        """Column id to which this cell belongs."""

    @property
    def style(self) -> ColumnStyle:
        """Style of column to which this cell belongs."""
        return self.data_table._column_styles[self.column_id]


class _ColumnLabel(_CellBase):
    """A column label cell in a data table."""

    def __init__(self, label: str, allow_sorting: bool, **kwargs):
        super().__init__(**kwargs)
        self._sort_state = _SortState.NOT_SORTED

        lines = label.split("\n")

        self.allow_sorting = allow_sorting
        self.label = label
        """Display text of cell."""
        self.cell_min_height = len(lines)
        """Minimum allowed height of cells."""
        self.cell_min_width = max(
            (
                max(wcswidth(line) for line in lines)
                + _SORT_INDICATOR_SPACING  # label width
                + _SORT_INDICATOR_WIDTH
                + 2 * self.style.padding
            ),
            self.style.min_width,
        )
        """Minimum allowed width of cell."""

    @property
    def indicator_pos(self) -> Point:
        return Point(
            self.height // 2, self.width - _SORT_INDICATOR_WIDTH - self.style.padding
        )

    @property
    def sort_state(self) -> _SortState:
        """Sorted state of column of which this label belongs."""
        return self._sort_state

    @sort_state.setter
    def sort_state(self, sort_state: _SortState):
        self._sort_state = _SortState(sort_state)
        if self.allow_sorting:
            self.canvas["ord"][self.indicator_pos] = egc_ord(self._sort_state.value)

    def _update_indicator(self):
        if self._allow_sorting:
            fg = self.data_table.get_color("data_table_sort_indicator_fg")
            bg = self.data_table.get_color("data_table_sort_indicator_bg")
            char = self._sort_state.value
        else:
            fg = self.data_table.get_color("primary_fg")
            bg = self.data_table.get_color("primary_bg")
            char = " "
        self.canvas["fg_color"][self.indicator_pos] = fg
        self.canvas["bg_color"][self.indicator_pos] = bg
        self.canvas["ord"][self.indicator_pos] = egc_ord(char)

    @property
    def allow_sorting(self) -> bool:
        return self._allow_sorting

    @allow_sorting.setter
    def allow_sorting(self, allow_sorting: bool):
        self._allow_sorting = allow_sorting and self.style.allow_sorting
        self._update_indicator()

    def on_size(self):
        super().on_size()
        self.clear()
        align = _ALIGN_FORMATTER[self.style.alignment]
        padding = self.style.padding
        content_width = (
            self.width - 2 * padding - _SORT_INDICATOR_WIDTH - _SORT_INDICATOR_SPACING
        )
        content = "\n".join(
            f"{line:{align}{content_width}}{' ' * _SORT_INDICATOR_SPACING}"
            for line in self.label.split("\n")
        )
        add_text(self.canvas[:, padding : -_SORT_INDICATOR_WIDTH - padding], content)
        self._update_indicator()

    def update_hover(self):
        self.data_table._update_hover()

    def on_release(self):
        if not self.allow_sorting:
            return

        if self.sort_state is _SortState.ASCENDING:
            self.sort_state = _SortState.DESCENDING
        else:
            self.sort_state = _SortState.ASCENDING

        for column_label in self.data_table._column_labels.children:
            if column_label is not self:
                column_label.sort_state = _SortState.NOT_SORTED

        self.data_table._sort(self.column_id, self.sort_state)


class _DataCell(_CellBase):
    """A data cell in a data table."""

    def __init__(
        self, data: SupportsLessThan, row_id: int, striped: bool = False, **kwargs
    ):
        super().__init__(**kwargs)

        self.data = data
        """Data of cell."""

        lines = self.style.render(data).split("\n")

        self.cell_min_height = len(lines)
        """Minimum allowed height of cell."""
        self.cell_min_width = max(
            (
                max(wcswidth(line) for line in lines)
                + 2 * self.style.padding  # width of rendered data
            ),
            self.style.min_width,
        )
        """Minimum allowed width of cell."""
        self.row_id = row_id
        """Row id of row cell belongs to."""
        self.striped = striped
        """Whether cell is striped."""
        self.selected = False
        """Whether cell is selected."""

    def on_size(self):
        super().on_size()
        self.chars[:] = " "
        if self.striped:
            fg = self.data_table.get_color("data_table_stripe_fg")
            bg = self.data_table.get_color("data_table_stripe_bg")
        elif self.selected:
            fg = self.data_table.get_color("data_table_selected_fg")
            bg = self.data_table.get_color("data_table_selected_bg")
        else:
            fg = self.data_table.get_color("primary_fg")
            bg = self.data_table.get_color("primary_bg")
        self.canvas["fg_color"] = fg
        self.canvas["bg_color"] = bg

        align = _ALIGN_FORMATTER[self.style.alignment]
        padding = self.style.padding
        content_width = self.width - 2 * padding
        content = "\n".join(
            f"{line:{align}{content_width}}"
            for line in self.style.render(self.data).split("\n")
        )
        add_text(self.canvas[:, padding : self.width - padding], content)

    def update_hover(self):
        self.data_table._update_hover(self.column_id, self.row_id)

    def on_release(self):
        self.data_table._on_release()


class _FauxPane(Pane):
    def _render(self, cells, graphics, kind):
        data_table: DataTable = self.parent.parent  # type: ignore
        self._region -= data_table._table._region
        super()._render(cells, graphics, kind)


class DataTable(Themable, Gadget):
    r"""
    A data table gadget.

    Parameters
    ----------
    data : dict[str, Sequence[T]] | None=None, default: None
        If given, construct a data table from this data. To gain more control over
        column styling use :meth:`add_column`.
    default_style : ColumnStyle | None, default: None
        Default style for new columns.
    select_items : Literal["cell", "row", "column"], default: "row"
        Determines which items are selected when data table is clicked.
    zebra_stripes : bool, default: True
        Whether alternate rows are colored differently.
    allow_sorting : bool, default: True
        Whether columns can be sorted.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    default_style : ColumnStyle
        Default style for new columns.
    select_items : Literal["cell", "row", "column"]
        Which items are selected when data table is clicked.
    zebra_stripes : bool
        Whether alternate rows are colored differently.
    allow_sorting : bool
        Whether columns can be sorted.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    add_column(label, ...)
        Add a column to the data table.
    add_row(data)
        Add a row to the data table.
    remove_column(column_id)
        Remove a column by column id.
    remove_row(row_id)
        Remove a row by row id.
    row_id_from_index(index)
        Returns the row id of the row at index.
    column_id_from_index(index)
        Returns the column id of the column at index.
    get_color(color_name)
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    _IDS = count()

    def __init__(
        self,
        *,
        data: dict[str, Sequence[SupportsLessThan]] | None = None,
        default_style: ColumnStyle | None = None,
        select_items: Literal["cell", "row", "column"] = "row",
        zebra_stripes: bool = True,
        allow_sorting: bool = True,
        alpha: float = 1.0,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._column_labels = _ColumnLabels(
            grid_rows=1, grid_columns=0, orientation="lr-tb", is_transparent=True
        )
        """Grid layout containing column label cells."""
        self._table = GridLayout(
            grid_rows=1, grid_columns=1, orientation="tb-lr", is_transparent=True
        )
        """Grid layout of column label and row grid layouts."""

        self._scroll_view = ScrollView(
            size_hint={"height_hint": 1.0, "width_hint": 1.0},
            dynamic_bars=True,
            is_transparent=is_transparent,
        )
        # Replace scroll view background with a pane that doesn't paint under _table.
        self._scroll_view.remove_gadget(self._scroll_view._background)
        self._scroll_view._background = _FauxPane(
            size_hint={"height_hint": 1.0, "width_hint": 1.0},
        )
        self._scroll_view.add_gadget(self._scroll_view._background)
        self._scroll_view.children.insert(0, self._scroll_view.children.pop())

        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self._column_ids: list[int] = []
        """Column ids. Index of id corresponds to index of column in table."""
        self._column_styles: dict[int, ColumnStyle] = {}
        """Column id to column style."""
        self._rows: dict[int, _DataCells] = {}
        """Row id to grid layout for row."""
        self._hover_column_id = -1
        """Column id that mouse is hovering or -1 if no columns are hovered."""
        self._hover_row_id = -1
        """Row id that mouse is hovering or -1 if no columns are hovered."""
        self.default_style = default_style or ColumnStyle()
        """Default style for new columns."""
        self._select_items: Literal["cell", "row", "column"] = select_items
        """Determines which items are selected when data table is clicked."""
        self._zebra_stripes = zebra_stripes
        """Whether alternate rows are colored differently."""
        self._allow_sorting = allow_sorting
        """Whether columns can be sorted."""
        self.alpha = alpha
        """Transparency of gadget."""

        self._table.add_gadget(self._column_labels)
        self._scroll_view.view = self._table
        self.add_gadget(self._scroll_view)

        if data is not None:
            for label, column_data in data.items():
                self.add_column(label, data=column_data)

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._scroll_view.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._scroll_view.alpha = alpha
        for child in self.walk():
            if isinstance(child, _CellBase):
                child.alpha = self._scroll_view.alpha

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._scroll_view.is_transparent = self.is_transparent
        for child in self.walk():
            if isinstance(child, _CellBase):
                child.is_transparent = self.is_transparent

    @property
    def select_items(self) -> Literal["cell", "row", "column"]:
        """Determines which items are selected when data table is clicked."""
        return self._select_items

    @select_items.setter
    def select_items(self, select_items: Literal["cell", "row", "column"]):
        select_items = select_items
        self._select_items = select_items
        for row in self._iter_rows():
            for cell in row.children:
                cell.selected = False
        self._repaint_cells()

    @property
    def zebra_stripes(self) -> bool:
        """Whether alternate rows are colored differently."""
        return self._zebra_stripes

    @zebra_stripes.setter
    def zebra_stripes(self, zebra_stripes: bool):
        self._zebra_stripes = zebra_stripes
        self._repaint_cells()

    @property
    def allow_sorting(self) -> bool:
        """Whether columns can be sorted."""
        return self._allow_sorting

    @allow_sorting.setter
    def allow_sorting(self, allow_sorting: bool):
        self._allow_sorting = allow_sorting
        for column_label in self._column_labels.children:
            column_label.allow_sorting = allow_sorting

    def update_theme(self):
        """Paint the gadget with current theme."""
        primary_fg = self.get_color("primary_fg")
        primary_bg = self.get_color("primary_bg")
        sort_fg = self.get_color("data_table_sort_indicator_fg")
        sort_bg = self.get_color("data_table_sort_indicator_bg")

        for label in self._column_labels.children:
            label.default_fg_color = primary_fg
            label.default_bg_color = primary_bg
            label.canvas["fg_color"] = primary_fg
            label.canvas["bg_color"] = primary_bg
            label.canvas["fg_color"][label.indicator_pos] = sort_fg
            label.canvas["bg_color"][label.indicator_pos] = sort_bg

        for row in self._rows.values():
            for cell in row.children:
                self._paint_cell_normal(cell)
        self._update_hover(self._hover_column_id, self._hover_row_id)

    def on_mouse(self, mouse_event: MouseEvent) -> bool | None:
        """Highlight row on mouse collision."""
        y, x = self.to_local(mouse_event.pos)
        if not (
            0 <= y < self._scroll_view.port_height
            and 0 <= x < self._scroll_view.port_width
        ):
            self._update_hover()
        return super().on_mouse(mouse_event)

    def _paint_cell_hover(self, cell: _DataCell):
        if cell.selected:
            fg = self.get_color("data_table_selected_hover_fg")
            bg = self.get_color("data_table_selected_hover_bg")
        elif cell.striped:
            fg = self.get_color("data_table_stripe_hover_fg")
            bg = self.get_color("data_table_stripe_hover_bg")
        else:
            fg = self.get_color("data_table_hover_fg")
            bg = self.get_color("data_table_hover_bg")
        cell.canvas["fg_color"] = fg
        cell.canvas["bg_color"] = bg

    def _paint_cell_normal(self, cell: _DataCell):
        if cell.selected:
            fg = self.get_color("data_table_selected_fg")
            bg = self.get_color("data_table_selected_bg")
        elif cell.striped:
            fg = self.get_color("data_table_stripe_fg")
            bg = self.get_color("data_table_stripe_bg")
        else:
            fg = self.get_color("primary_fg")
            bg = self.get_color("primary_bg")
        cell.canvas["fg_color"] = fg
        cell.canvas["bg_color"] = bg

    def _repaint_cells(self):
        for row in self._iter_rows():
            for i, cell in enumerate(row.children):
                cell.striped = self.zebra_stripes and i % 2 == 1
                self._paint_cell_normal(cell)

        column_id = self._hover_column_id
        row_id = self._hover_row_id
        self._hover_column_id = -1
        self._hover_row_id = -1
        self._update_hover(column_id, row_id)

    def _update_hover(self, column_id: int = -1, row_id: int = -1):
        if self.select_items == "row":
            if self._hover_row_id != row_id:
                if self._hover_row_id != -1:
                    for cell in self._rows[self._hover_row_id].children:
                        self._paint_cell_normal(cell)
                if row_id != -1:
                    for cell in self._rows[row_id].children:
                        self._paint_cell_hover(cell)
        elif self.select_items == "column":
            if self._hover_column_id != column_id:
                if self._hover_column_id != -1:
                    old_column_index = self._column_ids.index(self._hover_column_id)
                    for row in self._rows.values():
                        self._paint_cell_normal(row.children[old_column_index])
                if column_id != -1:
                    new_column_index = self._column_ids.index(column_id)
                    for row in self._rows.values():
                        self._paint_cell_hover(row.children[new_column_index])
        elif self.select_items == "cell":
            if self._hover_column_id != column_id or self._hover_row_id != row_id:
                if self._hover_column_id != -1 and self._hover_row_id != -1:
                    old_column_index = self._column_ids.index(self._hover_column_id)
                    self._paint_cell_normal(
                        self._rows[self._hover_row_id].children[old_column_index]
                    )
                if column_id != -1 and row_id != -1:
                    new_column_index = self._column_ids.index(column_id)
                    self._paint_cell_hover(
                        self._rows[row_id].children[new_column_index]
                    )

        self._hover_row_id = row_id
        self._hover_column_id = column_id

    def _on_release(self):
        if self.select_items == "row":
            for cell in self._rows[self._hover_row_id].children:
                cell.selected = not cell.selected
                self._paint_cell_hover(cell)
        elif self.select_items == "column":
            column_index = self._column_ids.index(self._hover_column_id)
            for row in self._rows.values():
                cell = row.children[column_index]
                cell.selected = not cell.selected
                self._paint_cell_hover(cell)
        elif self.select_items == "cell":
            column_index = self._column_ids.index(self._hover_column_id)
            cell = self._rows[self._hover_row_id].children[column_index]
            cell.selected = not cell.selected
            self._paint_cell_hover(cell)

    def _sort(self, column_id: int, sort_state: _SortState):
        column_index = self._column_ids.index(column_id)
        sorted_rows = sorted(
            self._iter_rows(),
            key=lambda row: row.children[column_index].data,
            reverse=sort_state is _SortState.DESCENDING,
        )
        self._table.children._gadgets = [self._table.children[0], *sorted_rows]
        self._table._invalidate_regions()
        self._table._reposition_children()

    def _fix_sizes(self):
        row_heights = [
            max(cell.cell_min_height for cell in row.children)
            for row in self._table.children
        ]
        column_widths = [
            max(row.children[i].cell_min_width for row in self._table.children)
            for i in range(len(self._column_ids))
        ]
        row: GridLayout
        cell: _CellBase
        for row, row_height in zip(self._table.children, row_heights):
            for cell, column_width in zip(row.children, column_widths):
                cell.size = row_height, column_width
            row.size = row.min_grid_size
        self._table.size = self._table.min_grid_size

        # Remove hovered rows/columns:
        self._hover_column_id = self._hover_row_id = -1
        self._repaint_cells()

    def _iter_rows(self) -> Iterator[_DataCells]:
        """Iterate over rows of table. The first row of column labels are skipped."""
        return islice(self._table.children, 1, None)

    def add_column(
        self,
        label: str,
        data: Sequence[SupportsLessThan] | None = None,
        style: ColumnStyle | None = None,
    ) -> int:
        """
        Add a column to the data table.

        If this is the first column added to the table, a row will be added for each
        item in `data`. Otherwise, the number of items in data must be equal to the
        number of rows in the table.

        Parameters
        ----------
        label : str
            The column label.
        data : Sequence[SupportsLessThan] | None, default: None
            Column data.
        style : ColumnStyle | None, default: None
            Column style. Uses :attr:`default_style` by default.

        Returns
        -------
        int
            Column id. This id can be used to remove the column.
        """
        if data is None:
            data = []
        if self._column_ids and len(data) != len(self._rows):
            raise ValueError(
                "Number of items in column data inconsistent with number of rows."
            )
        if style is None:
            style = replace(self.default_style)

        column_id = next(self._IDS)
        self._column_ids.append(column_id)
        self._column_styles[column_id] = style

        column_label = _ColumnLabel(
            data_table=self,
            column_id=column_id,
            label=label,
            allow_sorting=self.allow_sorting and style.allow_sorting,
            alpha=self.alpha,
            is_transparent=self.is_transparent,
        )
        self._column_labels.grid_columns += 1
        self._column_labels.add_gadget(column_label)

        if len(self._column_ids) == 1:
            self._table.grid_rows += len(data)
            for item in data:
                row_id = next(self._IDS)
                row = _DataCells(
                    grid_rows=1,
                    grid_columns=1,
                    orientation="lr-bt",
                    is_transparent=True,
                )
                row.add_gadget(
                    _DataCell(
                        data_table=self,
                        column_id=column_id,
                        data=item,
                        row_id=row_id,
                        alpha=self.alpha,
                        is_transparent=self.is_transparent,
                    )
                )
                self._rows[row_id] = row
                self._table.add_gadget(row)
        else:
            for item, row in zip(data, self._iter_rows()):
                row_id = row.children[0].row_id
                row.grid_columns += 1
                row.add_gadget(
                    _DataCell(
                        data_table=self,
                        column_id=column_id,
                        data=item,
                        row_id=row_id,
                        alpha=self.alpha,
                        is_transparent=self.is_transparent,
                    )
                )

        self._fix_sizes()
        return column_id

    def add_row(self, data: Sequence[SupportsLessThan]) -> int:
        """
        Add a row to the data table.

        There must be at least one column before a row can be added. The number
        of items in the row must match the number of columns.

        Parameters
        ----------
        data : Sequence[SupportsLessThan]
            The row data.

        Returns
        -------
        int
            Row id. This id can be used to remove the row.
        """
        if not self._column_ids or len(data) != len(self._column_ids):
            raise ValueError(
                "Number of items in row data inconsistent with number of columns."
            )

        row_id = next(self._IDS)
        row_layout = _DataCells(
            grid_rows=1,
            grid_columns=len(data),
            orientation="lr-bt",
            is_transparent=True,
        )
        row_layout.add_gadgets(
            _DataCell(
                data_table=self,
                column_id=column_id,
                data=item,
                row_id=row_id,
                alpha=self.alpha,
                is_transparent=self.is_transparent,
            )
            for column_id, item in zip(self._column_ids, data)
        )
        self._rows[row_id] = row_layout
        self._table.grid_rows += 1
        self._table.add_gadget(row_layout)
        self._fix_sizes()
        return row_id

    def remove_column(self, column_id: int):
        """
        Remove a column by column id.

        Column id can be retrieved by index with :meth:`column_id_from_index`.

        Parameters
        ----------
        column_id : int
            The id of the column to remove.
        """
        column_index = self._column_ids.index(column_id)
        del self._column_ids[column_index]
        for row in self._table.children:
            row.remove_gadget(row.children[column_index])
            row.grid_columns -= 1
        self._fix_sizes()

    def remove_row(self, row_id: int):
        """
        Remove a row by row id.

        Row id can be retrieved by index with :meth:`row_id_from_index`.

        Parameters
        ----------
        row_id : int
            The id of the row to remove.
        """
        row = self._rows.pop(row_id)
        self._table.remove_gadget(row)
        self._table.grid_rows -= 1
        self._fix_sizes()

    def row_id_from_index(self, index: int) -> int:
        """
        Return the row id of the row at index.

        Parameters
        ----------
        index : int
            Index of row in table.

        Returns
        -------
        int
            Row id of the row tat index.
        """
        return self._table.children[index + 1].children[0].row_id

    def column_id_from_index(self, index: int) -> int:
        """
        Return the column id of the column at index.

        Parameters
        ----------
        index : int
            Index of column in table.

        Returns
        -------
        int
            Column id of the column at index.
        """
        return self._column_labels.children[index].column_id



import asyncio
import os
import re
from collections.abc import Callable
from typing import Literal

from ..logging import get_logger
from .vt100_terminal cimport (
    CSI,
    CSI_PARAMS,
    DECRPM,
    ESCAPE,
    EXECUTE_NEXT,
    GROUND,
    OSC,
    PASTE,
    uint,
    uint8,
)
from ..colors import Color
from ..geometry import Size, Point
from ._fbuf cimport(
    fbuf,
    fbuf_endswith,
    fbuf_equals,
    fbuf_flush_fd,
    fbuf_free,
    fbuf_init,
    fbuf_put_char,
    fbuf_putn,
    fbuf_read_fd,
    fbuf_small_init,
)
from .ansi_escapes import ANSI_ESCAPES
from .events import (
    ColorReportEvent,
    CursorPositionReportEvent,
    DECReplyModeEvent,
    DeviceAttributesReportEvent,
    Event,
    FocusEvent,
    KeyEvent,
    MouseEvent,
    PasteEvent,
    PixelGeometryReportEvent,
    ResizeEvent,
    UnknownEscapeSequence,
)

DEF MAX_PARAMS = 32
DEF PASTE_START = b"\x1b[200~"
DEF PASTE_END = b"\x1b[201~"
DEF STRING_TERMINATOR = b"\x1b\\"
DEF FOCUS_IN = b"\x1b[I"
DEF FOCUS_OUT = b"\x1b[O"

ESCAPE_TIMEOUT: float = 0.05
DSR_REQUEST_TIMEOUT: float = 0.1
COLOR_RE: re.Pattern[bytes] = re.compile(
    rb"\x1b\]1([10]);rgb:"
    rb"([0-9a-f]{2})[0-9a-f]{2}/"
    rb"([0-9a-f]{2})[0-9a-f]{2}/"
    rb"([0-9a-f]{2})[0-9a-f]{2}\x1b\\"
)
DEC_MODES: frozenset[int] = frozenset([1016, 2026])

logger = get_logger(__name__)


cdef int csi_params(const fbuf *f, char *initial, char* final, uint *params):
    final[0] = f.buf[f.len - 1]
    if f.len == 3:
        return 0

    cdef size_t param_start = 2
    if not (0x30 <= f.buf[2] <= 0x39 or f.buf[2] == 0x3b):
        initial[0] = f.buf[2]
        if f.len == 4:
            return 0
        param_start += 1

    cdef size_t param_end
    if f.buf[f.len - 2] == 0x24:
        # DECRPM
        param_end = f.len - 2
    else:
        param_end = f.len - 1

    cdef:
        size_t i
        int nparams = 0
        uint param = 0

    for i in range(param_start, param_end):
        if f.buf[i] == 0x3b:
            params[nparams] = param
            nparams += 1
            if nparams == MAX_PARAMS:
                # Too many params.
                return -1
            param = 0
        else:
            param *= 10
            param += f.buf[i] - 0x30

    params[nparams] = param
    nparams += 1
    return nparams


cdef class Vt100Terminal:
    def __cinit__(self):
        if fbuf_small_init(&self.read_buf):
            raise MemoryError
        if fbuf_small_init(&self.in_buf):
            fbuf_free(&self.read_buf)
            raise MemoryError
        if fbuf_init(&self.out_buf):
            fbuf_free(&self.read_buf)
            fbuf_free(&self.in_buf)
            raise MemoryError

        self.state = GROUND
        self.last_y = 0
        self.last_x = 0
        self.skip_newline = 0
        self.sum_supported = 0
        self.sgr_pixels_mode = 0

    def __dealloc__(self):
        fbuf_free(&self.read_buf)
        fbuf_free(&self.in_buf)
        fbuf_free(&self.out_buf)

    def __init__(self, stdin: int = 0, stdout: int = 1):
        self.stdin = stdin
        self.stdout = stdout
        # FIXME: Should check if stdin/stdout is tty and read/flush appropriately.
        self.in_alternate_screen: bool = False
        self._event_buffer: list[Event] = []
        self._event_handler: Callable[[list[Event]], None] | None = None
        self._escape_timeout: asyncio.TimerHandle | None = None
        self._dsr_timeouts: dict[bytes, asyncio.TimerHandle] = {}

    cdef void add_event(self, event: Event):
        self.in_buf.len = 0
        self.state = GROUND
        self._event_buffer.append(event)
        if isinstance(event, UnknownEscapeSequence):
            logger.debug(event)
        else:
            logger.events(event)

    cdef void feed1(self, uint8 char_):
        if fbuf_put_char(&self.in_buf, char_):
            raise MemoryError
        cdef bytes data

        if self.state == OSC:
            if fbuf_endswith(&self.in_buf, STRING_TERMINATOR, 2):
                self.execute_osc()
        elif self.state == PASTE:
            # Gross \r handling...
            # Replace \r with \n, but skip \n if \r\n.
            if char_ == 0xd:
                self.in_buf.buf[self.in_buf.len - 1] = 0xa
                self.skip_newline = 1
            elif self.skip_newline:
                self.skip_newline = 0
                if char_ == 0xa:
                    self.in_buf.len -= 1
                    return
            if fbuf_endswith(&self.in_buf, PASTE_END, 6):
                data = self.in_buf.buf[:self.in_buf.len - 6]
                self.add_event(PasteEvent(data.decode()))
        elif char_ == 0x1b:
            self.state = ESCAPE
            self.in_buf.buf[0] = 0x1b
            self.in_buf.len = 1
        elif self.state == GROUND:
            if 0x20 <= char_ < 0x7f:
                data = <bytes>char_
                self.add_event(KeyEvent(data.decode()))
            else:
                self.execute_ansi_escapes()
        elif self.state == EXECUTE_NEXT:
            self.execute_ansi_escapes()
        elif self.state == DECRPM:
            self.execute_decrpm()
        elif self.state == ESCAPE:
            if char_ == 0x5b:
                self.state = CSI
            elif char_ == 0x4f:
                self.state = EXECUTE_NEXT
            elif char_ == 0x5d:
                self.state = OSC
            elif 0x20 <= char_ < 0x7f:
                data = <bytes>char_
                self.add_event(KeyEvent(data.decode(), alt=True))
            else:
                self.execute_ansi_escapes()
        elif self.state == CSI:
            if char_ == 0x5b:
                self.state = EXECUTE_NEXT
            elif 0x30 <= char_ <= 0x3f and char_ != 0x3a:
                self.state = CSI_PARAMS
            else:
                self.execute_csi()
        elif self.state is CSI_PARAMS:
            if char_ == 0x24:
                self.state = DECRPM
            elif char_ < 0x30 or char_ > 0x39 and char_ != 0x3b:
                self.execute_csi_params()

    cdef void execute_ansi_escapes(self):
        cdef bytes escape = self.in_buf.buf[:self.in_buf.len]
        if escape in ANSI_ESCAPES:
            self.add_event(KeyEvent(*ANSI_ESCAPES[escape]))
        else:
            self.add_event(UnknownEscapeSequence(escape))

    cdef void execute_csi(self):
        if fbuf_equals(&self.in_buf, FOCUS_IN, 3):
            self.add_event(FocusEvent("in"))
        elif fbuf_equals(&self.in_buf, FOCUS_OUT, 3):
            self.add_event(FocusEvent("out"))
        else:
            self.execute_ansi_escapes()

    cdef void execute_csi_params(self):
        if fbuf_equals(&self.in_buf, PASTE_START, 6):
            self.state = PASTE
            self.skip_newline = 0
            self.in_buf.len = 0
            return

        cdef:
            char initial = 0x0, final = 0x0
            uint[MAX_PARAMS] params
            int nparams = csi_params(&self.in_buf, &initial, &final, &params[0])

        cdef bytes escape
        if nparams < 0:
            escape = self.in_buf.buf[:self.in_buf.len]
            self.add_event(UnknownEscapeSequence(escape))
            return

        if self._dsr_timeouts:
            if initial == 0x0 and final == b"R" and nparams == 2:
                self.add_event(
                    CursorPositionReportEvent(Point(params[0] - 1, params[1] - 1))
                )
                timeout = self._dsr_timeouts.pop(b"\x1b[6n", None)
                if timeout is not None:
                    timeout.cancel()
                return

            if initial == b"?" and final == b"c":
                self.add_event(
                    DeviceAttributesReportEvent(
                        frozenset(int(param) for param in params[:nparams])
                    )
                )
                timeout = self._dsr_timeouts.pop(b"\x1b[c", None)
                if timeout is not None:
                    timeout.cancel()
                return

            if (
                initial == 0x0
                and final == b"t"
                and nparams == 3
                and (params[0] == 4 or params[0] == 6)
            ):
                self.add_event(
                    PixelGeometryReportEvent(
                        "terminal" if params[0] == 4 else "cell",
                        Size(params[1], params[2]),
                    )
                )
                if params[0] == 4:
                    timeout = self._dsr_timeouts.pop(b"\x1b[14t", None)
                else:
                    timeout = self._dsr_timeouts.pop(b"\x1b[16t", None)
                if timeout is not None:
                    timeout.cancel()
                return

        if initial == b"<" and (final == b"m" or final == b"M") and nparams == 3:
            self.execute_mouse(&params[0], final)
        else:
            self.execute_ansi_escapes()

    cdef void execute_mouse(self, uint *params, char state):
        cdef:
            uint info = params[0], x = params[1] - 1, y = params[2] - 1
            uint info_button = info % 4
            int dy = y - self.last_y, dx = x - self.last_x
            str button, event_type

        self.last_y = y
        self.last_x = x

        if info_button == 0:
            button = "left"
        elif info_button == 1:
            button = "middle"
        elif info_button == 2:
            button = "right"
        else:
            button = "no_button"

        if info & 64:
            if info & 1:
                event_type = "scroll_down"
            else:
                event_type = "scroll_up"
            button = "no_button"
        elif info & 32:
            event_type = "mouse_move"
        elif state == b"m":
            event_type = "mouse_up"
        elif button == "no_button":
            event_type = "mouse_move"
        else:
            event_type = "mouse_down"

        if self.sgr_pixels_mode:
            pos = None  # Will be set later in App using pixel geometry.
            pixels = Point(y, x)
        else:
            pos = Point(y, x)
            pixels = None

        self.add_event(
            MouseEvent(
                pos,
                button,
                event_type,
                bool(info & 8),
                bool(info & 16),
                bool(info & 4),
                dy,
                dx,
                pixels_pos=pixels,
            )
        )

    cdef void execute_osc(self):
        cdef:
            bytes escape = self.in_buf.buf[:self.in_buf.len]
            bytes kind, r, g, b

        if self._dsr_timeouts and (color := COLOR_RE.fullmatch(escape)):
            kind, r, g, b = color.groups()
            self.add_event(ColorReportEvent(
                "fg" if kind == b"0" else "bg",
                Color.from_hex(f"{r.decode()}{g.decode()}{b.decode()}"),
            ))
            if kind == b"0":
                timeout = self._dsr_timeouts.pop(b"\x1b]10;?\x1b\\", None)
            else:
                timeout = self._dsr_timeouts.pop(b"\x1b]11;?\x1b\\", None)
            if timeout is not None:
                timeout.cancel()
        else:
            self.add_event(UnknownEscapeSequence(escape))

    cdef void execute_decrpm(self):
        cdef:
            char initial = 0x0, final = 0x0
            uint[MAX_PARAMS] params
            int nparams = csi_params(&self.in_buf, &initial, &final, &params[0])
            uint mode = params[0], value = params[1]
            cdef bytes escape

        if nparams < 0 or not self._dsr_timeouts or mode not in DEC_MODES:
            escape = self.in_buf.buf[:self.in_buf.len]
            self.add_event(UnknownEscapeSequence(escape))
        else:
            timeout = self._dsr_timeouts.pop(b"\x1b[?%d$p" % mode, None)
            if timeout is not None:
                timeout.cancel()
            if mode == 2026:
                self.sum_supported = value

            self.add_event(DECReplyModeEvent(mode, value))

    def _timeout_escape(self) -> None:
        if self.state != PASTE:
            logger.events(
                "Timed out parsing ANSI escape. Expected if `escape` was pressed."
            )

            self.execute_ansi_escapes()

            if self._event_handler is not None:
                self._event_handler(self.events())
            return

        # Timed out during a paste. Check if PASTE_END was cutoff and remove it.
        logger.debug("Timed out during a paste.")

        cdef uint start
        if self.in_buf.len < 6:
            start = <uint>self.in_buf.len
        else:
            start = 6

        cdef uint i
        for i in range(start, 0, -1):
            if self.in_buf.buf[self.in_buf.len - i] == 0x1b:
                if fbuf_endswith(&self.in_buf, PASTE_END, i):
                    self.in_buf.len -= i
                break
        cdef bytes data = self.in_buf.buf[:self.in_buf.len]
        self.add_event(PasteEvent(data.decode()))

        if self._event_handler is not None:
            self._event_handler(self.events())

    cdef void dsr_request(self, escape: bytes):
        if escape in self._dsr_timeouts:
            # Already waiting for terminal response.
            return

        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            # Not monitoring stdin anyway.
            return

        self._dsr_timeouts[escape] = loop.call_later(
            DSR_REQUEST_TIMEOUT, self._create_dsr_timeout_callback(escape)
        )
        self.write(escape)
        self.flush()

    def _create_dsr_timeout_callback(self, escape: bytes) -> Callable[[], None]:
        def _on_timeout():
            if escape in self._dsr_timeouts:
                del self._dsr_timeouts[escape]
            logger.info("Device Status Report (DSR) Timeout: %s", escape)
        return _on_timeout

    cpdef void process_stdin(self):
        if self._escape_timeout is not None:
            self._escape_timeout.cancel()
            self._escape_timeout = None

        cdef int[2] size_event
        size_event[0] = -1
        cdef ssize_t read_result = fbuf_read_fd(
            &self.read_buf, self.stdin, &size_event[0]
        )
        if read_result < 0:
            raise OSError("Error reading fd.")
        elif read_result > 0:
            raise MemoryError

        if size_event[0] != -1:
            self.add_event(ResizeEvent(Size(size_event[0], size_event[1])))

        cdef size_t i
        for i in range(self.read_buf.len):
            self.feed1(self.read_buf.buf[i])
        self.read_buf.len = 0

        if self.state == GROUND:
            return

        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            pass
        else:
            self._escape_timeout = loop.call_later(ESCAPE_TIMEOUT, self._timeout_escape)

    def raw_mode(self) -> None:
        pass

    def restore_console(self) -> None:
        pass

    def attach(self, event_handler: Callable[[list[Event]], None]) -> None:
        pass

    def unattach(self) -> None:
        pass

    def feed(self, input_: bytes, reset_before: bool = True) -> list[Event]:
        if reset_before:
            if self._escape_timeout is not None:
                self._escape_timeout.cancel()
                self._escape_timeout = None

            self.state = GROUND
            self.in_buf.len = 0
            self._event_buffer.clear()

        for char_ in input_:
            self.feed1(char_)

        if self.state != GROUND:
            self.execute_ansi_escapes()

        return self.events()

    def events(self) -> list[Event]:
        events = self._event_buffer
        self._event_buffer = []
        return events

    def get_size(self) -> Size:
        cols, rows = os.get_terminal_size()
        return Size(rows, cols)

    def write(self, escape: bytes) -> None:
        if fbuf_putn(&self.out_buf, escape, len(escape)):
            raise MemoryError

    def flush(self) -> None:
        if fbuf_flush_fd(&self.out_buf, self.stdout):
            OSError("Failed to flush.")

    def set_title(self, title: str) -> None:
        self.write(b"\x1b]2;%b\x07" % title.encode())

    def enter_alternate_screen(self) -> None:
        self.write(b"\x1b[?1049h\x1b[H")
        self.in_alternate_screen = True

    def exit_alternate_screen(self) -> None:
        self.write(b"\x1b[?1049l")
        self.in_alternate_screen = False

    def enable_mouse_support(self) -> None:
        self.write(
            b"\x1b[?1000h"  # SET_VT200_MOUSE
            b"\x1b[?1003h"  # SET_ANY_EVENT_MOUSE
            b"\x1b[?1006h"  # SET_SGR_EXT_MODE_MOUSE
        )

    def disable_mouse_support(self) -> None:
        self.write(
            b"\x1b[?1000l"  # SET_VT200_MOUSE
            b"\x1b[?1003l"  # SET_ANY_EVENT_MOUSE
            b"\x1b[?1006l"  # SET_SGR_EXT_MODE_MOUSE
        )
        if self.sgr_pixels_mode:
            self.sgr_pixels_mode = 0
            self.write(b"\xb1[?1016l")  # SET_SGR_PIXELS_MODE_MOUSE

    def enable_sgr_pixels(self) -> None:
        self.sgr_pixels_mode = 1
        self.write(b"\xb1[?1016h")

    def disable_sgr_pixels(self) -> None:
        self.sgr_pixels_mode = 0
        self.write(b"\xb1[?1016l")

    def reset_attributes(self) -> None:
        self.write(b"\x1b[0m")

    def enable_bracketed_paste(self) -> None:
        self.write(b"\x1b[?2004h")

    def disable_bracketed_paste(self) -> None:
        self.write(b"\x1b[?2004l")

    def show_cursor(self) -> None:
        self.write(b"\x1b[?25h")

    def hide_cursor(self) -> None:
        self.write(b"\x1b[?25l")

    def enable_reporting_focus(self) -> None:
        self.write(b"\x1b[?1004h")

    def disable_reporting_focus(self) -> None:
        self.write(b"\x1b[?1004l")

    def request_cursor_position_report(self) -> None:
        self.dsr_request(b"\x1b[6n")

    def request_foreground_color(self) -> None:
        self.dsr_request(b"\x1b]10;?\x1b\\")

    def request_background_color(self) -> None:
        self.dsr_request(b"\x1b]11;?\x1b\\")

    def request_device_attributes(self) -> None:
        self.dsr_request(b"\x1b[c")

    def request_pixel_geometry(self) -> None:
        self.dsr_request(b"\x1b[16t")

    def request_terminal_geometry(self) -> None:
        self.dsr_request(b"\x1b[14t")

    def request_sgr_pixels_supported(self) -> None:
        # DECRQM 1016
        self.dsr_request(b"\x1b[?1016$p")

    def request_synchronized_update_mode_supported(self) -> None:
        # DECRQM 2026
        self.dsr_request(b"\x1b[?2026$p")

    def expect_dsr(self) -> bool:
        return bool(self._dsr_timeouts)

    def move_cursor(self, pos: Point) -> None:
        y, x = pos
        self.write(b"\x1b[%d;%dH" % (y + 1, x + 1))

    def erase_in_display(self, n: Literal[0, 1, 2, 3] = 0) -> None:
        self.write(b"\x1b[%dJ" % n)



"""A text-pad gadget for multiline editable text."""

from __future__ import annotations

from dataclasses import astuple, dataclass
from typing import Any, Literal

from numpy import ndarray
from tree_sitter import Tree
from ugrapheme import grapheme_iter, graphemes
from uwcwidth import wcswidth

from ..array_types import UInt32_2D
from ..colors import Neptune, SyntaxHighlightTheme
from ..geometry import clamp
from ..queries import Highlighter, TSPoint, changed_ranges_point_range, get_highlighter
from ..terminal.events import KeyEvent, MouseButton, MouseEvent, PasteEvent
from ..text_tools import canvas_as_text, egc_chr, egc_ord, is_word_char
from .behaviors.focusable import Focusable
from .behaviors.grabbable import Grabbable
from .behaviors.themable import Themable
from .cursor import Cursor
from .gadget import Gadget, Point, Pointlike, PosHint, Size, SizeHint, Sizelike
from .scroll_view import ScrollView
from .text import Text

__all__ = ["Point", "Size", "TextPad"]

EMPTY_LINE_CHARACTER = " "
"""Character used to display selected empty-lines."""


@dataclass(slots=True)
class _TextSelection:
    start: Point
    end: Point


@dataclass(slots=True)
class _TextEdit:
    cursor: Point
    selection: _TextSelection | None
    start: Point
    end: Point
    text: str

    @property
    def first(self) -> Point:
        return min(self.start, self.end)

    @first.setter
    def first(self, first: Point):
        if self.start <= self.end:
            self.start = first
        else:
            self.end = first

    @property
    def last(self) -> Point:
        return max(self.start, self.end)

    @last.setter
    def last(self, last: Point):
        if self.end >= self.start:
            self.end = last
        else:
            self.start = last

    def end_of_text(self) -> Point:
        y, x = self.first
        nlines = self.text.count("\n")
        if nlines:
            last_line = self.text[self.text.rfind("\n") + 1 :]
            return Point(y + nlines, wcswidth(last_line))
        return Point(y, x + wcswidth(self.text))

    def join(self, other: _TextEdit) -> bool:
        if other.end_of_text() == self.first:
            fy, fx = self.first
            ly, lx = self.last
            oly, olx = other.last
            self.first = other.first
            if ly == fy:
                self.last = Point(oly, olx + (lx - fx))
            else:
                self.last = Point(oly + (ly - fy), lx)
            self.text = other.text + self.text
            return True

        if self.last == other.first:
            self.last = other.last
            self.text += other.text
            return True

        return False


@dataclass
class _TreeEdit:
    start_byte_offset: int
    start_point: TSPoint
    end_byte_offset: int
    end_point: TSPoint


class TextPad(Themable, Grabbable, Focusable, Gadget):
    r"""
    A text-pad gadget for multiline editable text.

    Supports pasting, mouse selection, and cursor navigation.

    Parameters
    ----------
    syntax_highlight_language : str | None, default: None
        If provided, text will be syntax highlighted for given language.
    syntax_highlight_theme : SyntaxHighlightTheme, default: Neptune
        Color theme for syntax highlighting.
    alpha : float, default: 1.0
        Transparency of gadget.
    is_grabbable : bool, default: True
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool, default: False
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton | Literal["any"], default: "left"
        Mouse button used for grabbing.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    syntax_highlight_language : str | None
        If provided, text will be syntax highlighted for given language.
    syntax_highlight_theme : SyntaxHighlightTheme
        Color theme for syntax highlighting.
    alpha : float
        Transparency of gadget.
    text : str
        The text pad's text.
    cursor : Point
        The cursor position.
    page_lines : int
        Number of rows a page-up or -down moves.
    is_focused : bool
        Whether gadget has focus.
    any_focused : bool
        Whether any gadget has focus.
    is_grabbable : bool
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton | Literal["any"]
        Mouse button used for grabbing.
    is_grabbed : bool
        Whether gadget is grabbed.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    replace_text(start, end, text)
        Replace text from ``start`` to ``end`` with ``text``.
    undo()
        Undo previous edit.
    redo()
        Redo previous undo.
    move_cursor_left(n)
        Move cursor left `n` characters.
    move_cursor_right(n)
        Move cursor right `n` characters.
    move_cursor_up(n)
        Move cursor up `n` rows.
    move_cursor_down(n)
        Move cursor down `n` rows.
    move_word_left()
        Move cursor a word left.
    move_word_right()
        Move cursor a word right.
    get_color(color_name)
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    focus()
        Focus gadget.
    blur()
        Un-focus gadget.
    focus_next()
        Focus next focusable gadget.
    focus_previous()
        Focus previous focusable gadget.
    on_focus()
        Update gadget when it gains focus.
    on_blur()
        Update gadget when it loses focus.
    grab(mouse_event)
        Grab the gadget.
    ungrab(mouse_event)
        Ungrab the gadget.
    grab_update(mouse_event)
        Update gadget with incoming mouse events while grabbed.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        syntax_highlight_language: str | None = None,
        syntax_highlight_theme: SyntaxHighlightTheme = Neptune,
        alpha: float = 1.0,
        is_grabbable: bool = True,
        ptf_on_grab: bool = False,
        mouse_button: MouseButton | Literal["any"] = "left",
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ) -> None:
        self._cursor = Cursor()
        """The textpad's cursor."""
        self._prev_x: int | None = None
        """If the cursor has moved vertically, its previous x-coordinate."""
        self._pad = Text(size=(1, 1), is_transparent=is_transparent)
        """The textpad's text area."""
        self._scroll_view = ScrollView(
            size_hint={"height_hint": 1.0, "width_hint": 1.0},
            arrow_keys_enabled=False,
            is_grabbable=False,
            alpha=0,
            is_transparent=is_transparent,
            dynamic_bars=True,
        )
        self._selection: _TextSelection | None = None
        """Currently selected text."""
        self._active_line: int | None = None
        """If there is no selection, the active line is the line the cursor resides."""
        self._line_widths: list[int] = [0]
        """Line widths of each line in the text area."""
        self._replace_text_to_undo_stack: bool = True
        """Whether ``replace_text`` uses the undo or redo stack."""
        self._replace_text_clears_redo: bool = True
        """Whether ``replace_text`` clears the redo stack."""
        self._undo_stack: list[_TextEdit] = []
        """Stack of undo text edits."""
        self._redo_stack: list[_TextEdit] = []
        """Stack of redo text edits."""
        self._highlighter: Highlighter | None
        """Parser and queries for syntax highlighting."""
        if syntax_highlight_language is None:
            self._highlighter = None
        else:
            self._highlighter = get_highlighter(syntax_highlight_language)
        self._syntax_highlight_theme: SyntaxHighlightTheme = syntax_highlight_theme
        """Theme for syntax highlighting."""
        self._syntax_tree: Tree | None = None

        super().__init__(
            is_grabbable=is_grabbable,
            ptf_on_grab=ptf_on_grab,
            mouse_button=mouse_button,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.alpha = alpha
        self._pad.add_gadget(self._cursor)
        self._scroll_view.view = self._pad
        self.add_gadget(self._scroll_view)

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._pad.alpha

    @alpha.setter
    def alpha(self, alpha: float) -> None:
        self._pad.alpha = alpha

    @property
    def text(self) -> str:
        """The text pad's text."""
        return canvas_as_text(self._pad.canvas, self._line_widths)

    @text.setter
    def text(self, text: str):
        self._clear_selection()
        self.replace_text((0, 0), self.end_of_text, text)
        self._undo_stack.clear()
        self._redo_stack.clear()

    @property
    def cursor(self) -> Point:
        """The cursor position."""
        return self._cursor.pos

    @cursor.setter
    def cursor(self, cursor: Pointlike) -> None:
        self._cursor.pos = cursor
        self._scroll_view.scroll_to_rect(cursor)
        self._update_selection()
        self._update_active_line()

    @property
    def end_of_text(self) -> Point:
        """Point after last character in text."""
        return Point(len(self._line_widths) - 1, self._line_widths[-1])

    @property
    def page_lines(self) -> int:
        """Number of rows a page-up or page-down moves."""
        return self._scroll_view.port_height

    @property
    def syntax_highlight_language(self) -> str | None:
        """Language for syntax highlighting."""
        if self._highlighter is None:
            return None
        return self._highlighter.language_name

    @syntax_highlight_language.setter
    def syntax_highlight_language(self, syntax_highlight_language: str | None) -> None:
        if syntax_highlight_language is None:
            self._highlighter = None
        else:
            self._highlighter = get_highlighter(syntax_highlight_language)
        self.update_theme()

    def _clear_active_line(self) -> None:
        if self._active_line is None:
            return

        y = self._active_line
        self._active_line = None
        if self._highlighter is None:
            self._pad.canvas["fg_color"][y : y + 1] = self.get_color("primary_fg")
            self._pad.canvas["bg_color"][y : y + 1] = self.get_color("primary_bg")
        else:
            self._pad.canvas["bg_color"][y : y + 1] = (
                self._syntax_highlight_theme.default_bg
            )

    def _update_active_line(self) -> None:
        if self._selection is not None and self._selection.start != self._selection.end:
            return

        y = self.cursor.y
        if self._active_line is not None and self._active_line != y:
            self._clear_active_line()

        self._active_line = y

        if self._highlighter is None:
            self._pad.canvas["fg_color"][y : y + 1] = self.get_color(
                "text_pad_line_highlight_fg"
            )
            self._pad.canvas["bg_color"][y : y + 1] = self.get_color(
                "text_pad_line_highlight_bg"
            )
        else:
            self._pad.canvas["bg_color"][y : y + 1] = (
                self._syntax_highlight_theme.active_line
            )

    def _hide_empty_lines(self) -> None:
        """Replace any EMPTY_LINE_CHARACTER on empty lines with whitespace."""
        if self._selection is None:
            return

        if self._highlighter is None:
            fg_color = self.get_color("primary_fg")
            bg_color = self.get_color("primary_bg")
        else:
            fg_color = self._syntax_highlight_theme.default_fg
            bg_color = self._syntax_highlight_theme.default_bg

        sy, ey = self._selection.start.y, self._selection.end.y
        if sy > ey:
            sy, ey = ey, sy
        for y in range(sy, ey):
            if self._line_widths[y] == 0:
                self._pad.canvas[y, 0:1]["ord"] = egc_ord(" ")
                self._pad.canvas[y, 0:1]["fg_color"] = fg_color
                self._pad.canvas[y, 0:1]["bg_color"] = bg_color

    def _show_empty_lines(self) -> None:
        """Show empty lines in selection with EMPTY_LINE_CHARACTER."""
        if self._selection is None:
            return

        sy, ey = self._selection.start.y, self._selection.end.y
        if sy > ey:
            sy, ey = ey, sy

        if self._highlighter is None:
            fg_color = self.get_color("text_pad_selection_highlight_fg")
            bg_color = self.get_color("text_pad_selection_highlight_bg")
        else:
            fg_color = self._syntax_highlight_theme.default_fg
            bg_color = self._syntax_highlight_theme.selection

        for y in range(sy, ey):
            if self._line_widths[y] == 0 and y != self._cursor.y:
                self._pad.canvas[y, 0:1]["ord"] = egc_ord(EMPTY_LINE_CHARACTER)
                self._pad.canvas[y, 0:1]["fg_color"] = fg_color
                self._pad.canvas[y, 0:1]["bg_color"] = bg_color

    def _fill_array(self, start: Point, end: Point, arr: ndarray, value: Any) -> None:
        """Fill ``arr`` from ``start`` to ``end`` with ``value``."""
        if start < end:
            (sy, sx), (ey, ex) = start, end
        else:
            (sy, sx), (ey, ex) = end, start

        if sy == ey:
            arr[sy, sx:ex] = value
        else:
            arr[sy, sx : self._line_widths[sy]] = value
            for i in range(sy + 1, ey):
                arr[i, : self._line_widths[i]] = value
            arr[ey, :ex] = value

    def _repaint_selection(self, start: Point, end: Point, *, selected: bool) -> None:
        """Repaint text from ``start`` to ``end`` as if it was selected or not."""
        if start == end:
            return

        if selected:
            if self._highlighter is None:
                fg = self.get_color("text_pad_selection_highlight_fg")
                bg = self.get_color("text_pad_selection_highlight_bg")
            else:
                fg = None
                bg = self._syntax_highlight_theme.selection
        else:
            if self._highlighter is None:
                fg = self.get_color("primary_fg")
                bg = self.get_color("primary_bg")
            else:
                fg = None
                bg = self._syntax_highlight_theme.default_bg

        if fg is not None:
            self._fill_array(start, end, self._pad.canvas["fg_color"], fg)
        self._fill_array(start, end, self._pad.canvas["bg_color"], bg)

    def _clear_selection(self) -> None:
        if self._selection is None:
            return

        self._hide_empty_lines()
        start = self._selection.start
        end = self._selection.end
        self._selection = None
        self._repaint_selection(start, end, selected=False)

    def _update_selection(self) -> None:
        """
        Expand or shrink and repaint currently selected text by setting selection.end to
        current cursor position.
        """
        if self._selection is None:
            return

        if self._active_line is not None:
            self._clear_active_line()
        self._hide_empty_lines()

        if (
            self.cursor <= self._selection.start <= self._selection.end
            or self._selection.end <= self._selection.start <= self.cursor
        ):
            clear_start = self._selection.start
            clear_end = self._selection.end
            paint_start = self.cursor
            paint_end = self._selection.start
        elif (
            self.cursor <= self._selection.end <= self._selection.start
            or self._selection.start <= self._selection.end <= self.cursor
        ):
            clear_start = None
            clear_end = None
            paint_start = self.cursor
            paint_end = self._selection.end
        else:
            # self._selection.start <= self.cursor <= self._selection.end
            # or self._selection.end <= self.cursor <= self._selection.start
            clear_start = self.cursor
            clear_end = self._selection.end
            paint_start = None
            paint_end = None

        self._selection.end = self.cursor
        if clear_start is not None and clear_end is not None:
            self._repaint_selection(clear_start, clear_end, selected=False)
        if paint_start is not None and paint_end is not None:
            self._repaint_selection(paint_start, paint_end, selected=True)

        self._show_empty_lines()

    def _tree_edit(self, start: Pointlike, end: Pointlike) -> _TreeEdit | None:
        """
        Convert point (row, column) in text pad to tree-sitter points and byte
        offsets.
        """
        if self._highlighter is None:
            return None

        ords: UInt32_2D = self._pad.canvas["ord"]

        sy, sx = start
        start_byte_offset = 0
        start_row_byte_offset = 0

        for v in range(sy):
            for u in range(self._line_widths[v]):
                start_byte_offset += len(egc_chr(ords[v, u]).encode())
            start_byte_offset += 1

        for u in range(sx):
            start_row_byte_offset += len(egc_chr(ords[sy, u]).encode())

        start_byte_offset += start_row_byte_offset

        ey, ex = end
        nbytes = ey - sy
        if sy == ey:
            for u in range(sx, ex):
                nbytes += len(egc_chr(ords[sy, u]).encode())
            end_row_byte_offset = start_row_byte_offset + nbytes
        else:
            # Last bit of first line
            for u in range(sx, self._line_widths[sy]):
                nbytes += len(egc_chr(ords[sy, u]).encode())

            # All of middle lines
            for v in range(sy + 1, ey):
                for u in range(self._line_widths[v]):
                    nbytes += len(egc_chr(ords[v, u]).encode())

            # First bit of last line
            end_row_byte_offset = 0
            for u in range(ex):
                end_row_byte_offset += len(egc_chr(ords[ey, u]).encode())
            nbytes += end_row_byte_offset

        end_byte_offset = start_byte_offset + nbytes
        return _TreeEdit(
            start_byte_offset,
            (sy, start_row_byte_offset),
            end_byte_offset,
            (ey, end_row_byte_offset),
        )

    def _pos_to_point(self, pos: Point, point: Pointlike, byte_offset: int) -> TSPoint:
        """
        Convert ``pos`` (row, column) to a tree sitter point (row, byte_offset).

        The return value is calculated starting from previously calculated ``point``
        and ``byte_offset`` to save some effort.
        """
        ords: UInt32_2D = self._pad.canvas["ord"]
        sy, sx = point
        ey, ex = pos
        if ey == sy:
            for u in range(sx, ex):
                byte_offset += len(egc_chr(ords[sy, u]).encode())
            return ey, byte_offset

        byte_offset = 0
        for u in range(ex):
            byte_offset += len(egc_chr(ords[ey, u]).encode())
        return ey, byte_offset

    def _tree_sitter_read_pad(self, _, point: tuple[int, int]) -> bytes:
        y, x = self._pad._tree_sitter_point_to_pos(point)
        if y >= len(self._line_widths):
            return b""

        end_column = self._line_widths[y]
        if self._pad._injection_range is not None:
            if y > self._pad._injection_range.end_row:
                return b""

            if y == self._pad._injection_range.end_row:
                end_column = self._pad._injection_range.end_column

        ords = self._pad.canvas["ord"][y, x:end_column].tolist()
        return "".join(egc_chr(ord_) for ord_ in ords).encode() + b"\n"

    def _highlight(
        self, tree_edit: _TreeEdit, new_end_byte: int, new_end_point: TSPoint
    ) -> None:
        """Incremental highlight of text."""
        if self._highlighter is None:
            return None

        old_tree = self._syntax_tree
        if old_tree is None:
            self._syntax_tree = self._highlighter.parser.parse(
                self._tree_sitter_read_pad
            )
            return

        old_tree.edit(
            tree_edit.start_byte_offset,
            tree_edit.end_byte_offset,
            new_end_byte,
            tree_edit.start_point,
            tree_edit.end_point,
            new_end_point,
        )
        self._syntax_tree = self._highlighter.parser.parse(
            self._tree_sitter_read_pad, old_tree
        )

        changed_ranges = old_tree.changed_ranges(self._syntax_tree)
        if changed_ranges:
            point_range = changed_ranges_point_range(changed_ranges)
        else:
            point_range = tree_edit.start_point, new_end_point

        self._pad._highlight(
            self._syntax_highlight_theme,
            self._highlighter,
            self._syntax_tree,
            point_range,
        )

    def replace_text(self, start: Pointlike, end: Pointlike, text: str) -> None:
        """
        Replace text pad text from ``start`` to ``end`` with ``text``.

        Parameters
        ----------
        start : Pointlike
            The start of text to replace.
        end : Pointlike
            The end of text to replace.
        text : str
            The replacement text.
        """
        prev_selection = self._selection
        prev_cursor = self.cursor

        self._clear_selection()
        self._clear_active_line()
        self._prev_x = None

        pad = self._pad
        nlines = len(self._line_widths)

        if start > end:
            start, end = end, start

        tree_edit = self._tree_edit(start, end)

        sy, sx = start
        ey, ex = end

        # Clamp start and end within text.
        sy = clamp(sy, 0, nlines - 1)
        sx = clamp(sx, 0, self._line_widths[sy])
        ey = clamp(ey, 0, nlines - 1)
        ex = clamp(ex, 0, self._line_widths[ey])

        if sy == ey:
            replaced = canvas_as_text(pad.canvas[sy, sx:ex])
        else:
            replaced_lines = [
                canvas_as_text(pad.canvas[sy, sx : self._line_widths[sy]])
            ]
            for y in range(sy + 1, ey):
                replaced_lines.append(
                    canvas_as_text(pad.canvas[y, : self._line_widths[y]])
                )
            replaced_lines.append(canvas_as_text(pad.canvas[ey, :ex]))
            replaced = "\n".join(replaced_lines)

        line_after_replace = pad.canvas[ey, ex : self._line_widths[ey]].copy()
        line_after_replace_width = self._line_widths[ey] - ex

        lines = text.split("\n")
        new_lines = len(lines) - 1

        line_widths: list[int] = [wcswidth(line) for line in lines]
        line_widths[0] += sx
        end_x = line_widths[-1]
        line_widths[-1] += line_after_replace_width
        self._line_widths[sy : ey + 1] = line_widths

        # Label all lines after ``end`` "remainder":
        # If the pad will shrink, move remainder before resizing pad (before
        # remainder is clipped). Otherwise, move remainder after resizing pad (so pad
        # has enough vertical space).
        dy = new_lines - ey + sy
        if dy < 0:
            pad.canvas[ey + 1 + dy : nlines + dy] = pad.canvas[ey + 1 : nlines]
            pad.canvas[nlines + dy :] = pad.default_cell

        self._set_pad_size()

        if dy > 0:
            pad.canvas[ey + 1 + dy : nlines + dy] = pad.canvas[ey + 1 : nlines]

        end_y = sy + new_lines
        if new_lines == 0:
            [line] = lines
            pad.add_str(line, pos=start)
            pad.canvas[sy, end_x : self._line_widths[sy]] = line_after_replace
            pad.canvas[sy, self._line_widths[sy] :] = pad.default_cell
        else:
            first_line, *middle_lines, last_line = lines
            pad.add_str(first_line, pos=start)
            pad.canvas[sy, self._line_widths[sy] :] = pad.default_cell
            for y, line in enumerate(middle_lines, start=sy + 1):
                pad.add_str(line, pos=(y, 0))
                pad.canvas[y, self._line_widths[y] :] = pad.default_cell
            pad.add_str(last_line, pos=(end_y, 0))
            pad.canvas[
                end_y,
                end_x : self._line_widths[end_y],
            ] = line_after_replace
            pad.canvas[end_y, self._line_widths[end_y] :] = pad.default_cell
        cursor = Point(end_y, end_x)

        inverse_edit = _TextEdit(
            prev_cursor, prev_selection, Point(sy, sx), cursor, replaced
        )
        if self._replace_text_to_undo_stack:
            stack = self._undo_stack
        else:
            stack = self._redo_stack
            self._replace_text_to_undo_stack = True

        if self._replace_text_clears_redo:
            self._redo_stack.clear()
        else:
            self._replace_text_clears_redo = True

        if not (stack and stack[-1].join(inverse_edit)):
            stack.append(inverse_edit)

        if tree_edit is not None:
            new_end_byte = tree_edit.start_byte_offset + len(text.encode())
            new_end_point = self._pos_to_point(cursor, start, tree_edit.start_point[1])
            self._highlight(tree_edit, new_end_byte, new_end_point)

        self.cursor = cursor

    def undo(self) -> None:
        """Undo previous edit."""
        if not self._undo_stack:
            return

        last_edit = self._undo_stack.pop()
        self._replace_text_to_undo_stack = False
        self._replace_text_clears_redo = False
        self.replace_text(last_edit.start, last_edit.end, last_edit.text)
        self.cursor = last_edit.cursor

        if last_edit.selection is not None:
            self._selection = last_edit.selection
            self._clear_active_line()
            self._repaint_selection(
                self._selection.start, self._selection.end, selected=True
            )
            self._show_empty_lines()

    def redo(self) -> None:
        """Redo previous undo."""
        if not self._redo_stack:
            return

        prev_edit = self._redo_stack.pop()
        self._replace_text_clears_redo = False
        self.replace_text(prev_edit.start, prev_edit.end, prev_edit.text)
        self.cursor = prev_edit.cursor

        if prev_edit.selection is not None:
            self._selection = prev_edit.selection
            self._clear_active_line()
            self._repaint_selection(
                self._selection.start, self._selection.end, selected=True
            )
            self._show_empty_lines()

    def move_cursor_left(self, n: int = 1) -> None:
        """Move cursor left `n` characters."""
        self._prev_x = None
        y, x = self._cursor.pos

        while n > 0:
            text_before_cursor = canvas_as_text(self._pad.canvas[y, :x])
            egcs = graphemes(text_before_cursor)
            if n <= len(egcs):
                x = wcswidth(egcs[:-n])
                break

            if y == 0:
                x = 0
                break

            y -= 1
            x = self._line_widths[y]
            n -= len(egcs) + 1

        self.cursor = y, x

    def move_cursor_right(self, n: int = 1) -> None:
        """Move cursor right `n` characters."""
        self._prev_x = None
        y, x = self._cursor.pos

        while n > 0:
            text_after_cursor = canvas_as_text(
                self._pad.canvas[y, x : self._line_widths[y]]
            )
            egcs = graphemes(text_after_cursor)
            if n <= len(egcs):
                x += wcswidth(egcs[:n])
                break

            if y == self.end_of_text.y:
                x = self._line_widths[y]
                break

            y += 1
            n -= len(egcs) + 1
            x = 0

        self.cursor = y, x

    def _fix_x(self, y, x) -> int:
        line = canvas_as_text(self._pad.canvas[y, : self._line_widths[y]])
        current_x = 0
        for egc in grapheme_iter(line):
            if current_x + wcswidth(egc) > x:
                return current_x
            current_x += wcswidth(egc)
        return self._line_widths[y]

    def move_cursor_up(self, n: int = 1) -> None:
        """Move cursor up `n` rows."""
        y, x = self._cursor.pos

        if self._prev_x is None or y == x == 0:
            self._prev_x = x

        if y > 0:
            y = max(0, y - n)
            x = self._fix_x(y, min(self._prev_x, self._line_widths[y]))
        else:
            x = 0

        self.cursor = y, x

    def move_cursor_down(self, n: int = 1) -> None:
        """Move cursor down `n` rows."""
        y, x = self._cursor.pos
        ey, ex = self.end_of_text

        if self._prev_x is None or y == ey and x == ex:
            self._prev_x = x

        if y < ey:
            y = min(ey, y + n)
            x = self._fix_x(y, min(self._prev_x, self._line_widths[y]))
        else:
            x = ex

        self.cursor = y, x

    def move_word_left(self) -> None:
        """Move cursor a word left."""
        self._prev_x = None
        last_x = self.cursor.x
        first_char_found = char_is_word_char = False
        while True:
            self.move_cursor_left()
            if self.cursor.x == last_x:
                break

            last_x = self.cursor.x

            current_char = egc_chr(self._pad.canvas[self.cursor]["ord"])
            if not first_char_found:
                if not current_char.isspace():
                    first_char_found = True
                    char_is_word_char = is_word_char(current_char)
            elif current_char.isspace() or char_is_word_char != is_word_char(
                current_char
            ):
                self.move_cursor_right()
                break

    def move_word_right(self) -> None:
        """Move cursor a word right."""
        self._prev_x = None
        last_x = self.cursor.x
        first_char_found = char_is_word_char = False
        while True:
            self.move_cursor_right()
            if self.cursor.x == last_x:
                break

            last_x = self.cursor.x

            current_char = egc_chr(self._pad.canvas[self.cursor]["ord"])
            if not first_char_found:
                if not current_char.isspace():
                    first_char_found = True
                    char_is_word_char = is_word_char(current_char)
            elif current_char.isspace() or char_is_word_char != is_word_char(
                current_char
            ):
                break

    def _enter(self) -> None:
        if self._selection is None:
            start = end = self.cursor
        else:
            start, end = self._selection.start, self._selection.end
        self.replace_text(start, end, "\n")
        self._redo_stack.clear()

    def _tab(self) -> None:
        if self._selection is None:
            start = end = self.cursor
        else:
            start, end = self._selection.start, self._selection.end
        self.replace_text(start, end, "    ")
        self._redo_stack.clear()

    def _backspace(self) -> None:
        if self._selection is None:
            end = self.cursor
            self.move_cursor_left()
            start = self.cursor
            self.cursor = end
            if start == end:
                return
        else:
            start, end = self._selection.start, self._selection.end
        self.replace_text(start, end, "")

    def _delete(self) -> None:
        if self._selection is None:
            start = self.cursor
            self.move_cursor_right()
            end = self.cursor
            self.cursor = start
            if start == end:
                return
        else:
            start, end = self._selection.start, self._selection.end
        self.replace_text(start, end, "")

    def _left(self) -> None:
        if self._selection is None:
            self.move_cursor_left()
        else:
            select_start = min(self._selection.start, self._selection.end)
            self._clear_selection()
            self.cursor = select_start

    def _right(self) -> None:
        if self._selection is None:
            self.move_cursor_right()
        else:
            select_end = max(self._selection.start, self._selection.end)
            self._clear_selection()
            self.cursor = select_end

    def _ctrl_left(self) -> None:
        self._clear_selection()
        self.move_word_left()

    def _ctrl_right(self) -> None:
        self._clear_selection()
        self.move_word_right()

    def _ctrl_a(self) -> None:
        """Select all."""
        self._selection = _TextSelection(Point(0, 0), self.end_of_text)
        self._repaint_selection(
            self._selection.start, self._selection.end, selected=True
        )
        self.cursor = self.end_of_text

    def _ctrl_d(self) -> None:
        """Select word."""
        self._clear_selection()
        last_x = self.cursor.x
        while True:
            self.move_cursor_left()
            if last_x == self.cursor.x:
                break
            if not is_word_char(egc_chr(self._pad.canvas[self.cursor]["ord"])):
                self.move_cursor_right()
                break
            last_x = self.cursor.x

        self._selection = _TextSelection(self.cursor, self.cursor)
        last_x = self.cursor.x
        while True:
            if not is_word_char(egc_chr(self._pad.canvas[self.cursor]["ord"])):
                break
            self.move_cursor_right()
            if last_x == self.cursor.x:
                break
            last_x = self.cursor.x

    def _up(self) -> None:
        if self._selection is not None:
            select_start = min(self._selection.start, self._selection.end)
            self._clear_selection()
            self.cursor = select_start
        self.move_cursor_up()

    def _down(self) -> None:
        if self._selection is not None:
            select_end = max(self._selection.start, self._selection.end)
            self._clear_selection()
            self.cursor = select_end
        self.move_cursor_down()

    def _pgup(self) -> None:
        if self._selection is not None:
            select_start = min(self._selection.start, self._selection.end)
            self._clear_selection()
            self.cursor = select_start
        self.move_cursor_up(self.page_lines)

    def _pgdn(self) -> None:
        if self._selection is not None:
            select_end = max(self._selection.start, self._selection.end)
            self._clear_selection()
            self.cursor = select_end
        self.move_cursor_down(self.page_lines)

    def _home(self) -> None:
        self._clear_selection()
        self._prev_x = None
        self.cursor = self.cursor.y, 0

    def _end(self) -> None:
        self._clear_selection()
        self._prev_x = None
        y = self.cursor.y
        self.cursor = y, self._line_widths[y]

    def _shift_left(self) -> None:
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self.move_cursor_left()

    def _shift_right(self) -> None:
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self.move_cursor_right()

    def _shift_ctrl_left(self) -> None:
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self.move_word_left()

    def _shift_ctrl_right(self) -> None:
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self.move_word_right()

    def _shift_up(self) -> None:
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self.move_cursor_up()

    def _shift_down(self) -> None:
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self.move_cursor_down()

    def _shift_pgup(self) -> None:
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self.move_cursor_up(self.page_lines)

    def _shift_pgdn(self) -> None:
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self.move_cursor_down(self.page_lines)

    def _shift_home(self) -> None:
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self._prev_x = None
        self.cursor = self.cursor.y, 0

    def _shift_end(self) -> None:
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self._prev_x = None
        y = self.cursor.y
        self.cursor = y, self._line_widths[y]

    def _escape(self) -> None:
        if self._selection is None:
            self.blur()
        else:
            self._clear_selection()

    def _ascii(self, key: str) -> None:
        if self._selection is None:
            start = end = self.cursor
        else:
            start, end = self._selection.start, self._selection.end
        self.replace_text(start, end, key)

    __HANDLERS = {
        ("enter", False, False, False): _enter,
        ("tab", False, False, False): _tab,
        ("backspace", False, False, False): _backspace,
        ("delete", False, False, False): _delete,
        ("left", False, False, False): _left,
        ("right", False, False, False): _right,
        ("left", False, True, False): _ctrl_left,
        ("right", False, True, False): _ctrl_right,
        ("up", False, False, False): _up,
        ("down", False, False, False): _down,
        ("page_up", False, False, False): _pgup,
        ("page_down", False, False, False): _pgdn,
        ("home", False, False, False): _home,
        ("end", False, False, False): _end,
        ("left", False, False, True): _shift_left,
        ("right", False, False, True): _shift_right,
        ("left", False, True, True): _shift_ctrl_left,
        ("right", False, True, True): _shift_ctrl_right,
        ("up", False, False, True): _shift_up,
        ("down", False, False, True): _shift_down,
        ("page_up", False, False, True): _shift_pgup,
        ("page_down", False, False, True): _shift_pgdn,
        ("home", False, False, True): _shift_home,
        ("end", False, False, True): _shift_end,
        ("escape", False, False, False): _escape,
        ("z", False, True, False): undo,
        ("y", False, True, False): redo,
        ("r", False, True, False): redo,
        ("a", False, True, False): _ctrl_a,
        ("d", False, True, False): _ctrl_d,
    }

    def on_key(self, key_event: KeyEvent) -> bool | None:
        """Process key press."""
        if not self.is_focused:
            return super().on_key(key_event)

        if (
            not key_event.alt
            and not key_event.ctrl
            and not key_event.shift
            and len(key_event.key) == 1
        ):
            self._ascii(key_event.key)
        elif handler := self.__HANDLERS.get(
            astuple(key_event)
        ):  # FIXME: Might break if extra args added to key_event
            handler(self)
        else:
            return super().on_key(key_event)

        return True

    def on_paste(self, paste_event: PasteEvent) -> bool | None:
        """Add paste to text pad."""
        if not self.is_focused:
            return

        if self._selection is None:
            start = end = self.cursor
        else:
            start, end = self._selection.start, self._selection.end
        self.replace_text(start, end, paste_event.paste)
        self._redo_stack.clear()

        return True

    def grab(self, mouse_event) -> None:
        """Start selection on grab."""
        if self._pad.collides_point(mouse_event.pos):
            super().grab(mouse_event)

            y, x = self._pad.to_local(mouse_event.pos)
            if y >= len(self._line_widths):
                return

            x = min(x, self._line_widths[y])
            cursor = Point(y, x)

            if not mouse_event.shift or self._selection is None:
                self._clear_selection()
                self._selection = _TextSelection(cursor, cursor)

            self.cursor = cursor

    def grab_update(self, mouse_event: MouseEvent) -> None:
        """Update selection on grab update."""
        if self._pad.collides_point(mouse_event.pos):
            y, x = self._pad.to_local(mouse_event.pos)
            if y < len(self._line_widths):
                x = min(x, self._line_widths[y])
                self.cursor = y, x
        else:
            cy, cx = self.cursor
            y, x = self.to_local(mouse_event.pos)
            h, w = self.size

            if y < 0:
                self.move_cursor_up()
            elif y >= h:
                self.move_cursor_down()

            if x < 0:
                if cx > 0:
                    self.move_cursor_left()
            elif x >= w:
                if cx < self._line_widths[cy]:
                    self.move_cursor_right()

    def ungrab(self, mouse_event) -> None:
        """Clear an empty selection on ungrab."""
        super().ungrab(mouse_event)
        if self._selection is not None and self._selection.start == self._selection.end:
            self._clear_selection()

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._pad.is_transparent = self.is_transparent
        self._scroll_view.is_transparent = self.is_transparent

    def update_theme(self) -> None:
        """Paint the gadget with current theme."""
        selection = self._selection
        self._clear_selection()

        if self._highlighter is None:
            self._cursor.reverse = True
            self._cursor.fg_color = None
            self._cursor.bg_color = None
            self._pad.default_fg_color = self.get_color("primary_fg")
            self._pad.default_bg_color = self.get_color("primary_bg")
            self._pad.canvas[["style", "fg_color", "bg_color"]] = (
                self._pad.default_cell[["style", "fg_color", "bg_color"]]
            )
            self._syntax_tree = None
        else:
            self._cursor.reverse = None
            self._cursor.fg_color = self._syntax_highlight_theme.cursor_fg
            self._cursor.bg_color = self._syntax_highlight_theme.cursor_bg
            self._pad.default_fg_color = self._syntax_highlight_theme.default_fg
            self._pad.default_bg_color = self._syntax_highlight_theme.default_bg
            self._syntax_tree = self._highlighter.parser.parse(
                self._tree_sitter_read_pad
            )
            self._pad._highlight(
                self._syntax_highlight_theme, self._highlighter, self._syntax_tree
            )

        self._selection = selection
        if selection is None or selection.start == selection.end:
            self._update_active_line()
        else:
            self._repaint_selection(selection.start, selection.end, selected=True)
            self._show_empty_lines()

    def _set_pad_size(self) -> None:
        nlines = len(self._line_widths)
        pad_width = max(self._line_widths) + 1
        h, w = self.size
        height = max(nlines, h - (pad_width > w))
        width = max(pad_width, w - 2 * (nlines > h))
        self._pad.size = height, width

    def on_size(self) -> None:
        """Update pad size on resize."""
        self._set_pad_size()

    def on_focus(self) -> None:
        """Show cursor on focus."""
        self._cursor.is_enabled = True
        self._update_active_line()

    def on_blur(self) -> None:
        """Hide cursor on blur."""
        self._cursor.is_enabled = False
        self._clear_selection()
        self._clear_active_line()



"""A bar chart gadget."""

from uwcwidth import wcswidth

from ..colors import NEPTUNE_PRIMARY_BG, NEPTUNE_PRIMARY_FG, Color, rainbow_gradient
from ..text_tools import Style, add_text, smooth_vertical_bar
from .gadget import Gadget, Point, Pointlike, PosHint, Size, SizeHint, Sizelike, lerp
from .pane import Pane
from .scroll_view import ScrollView
from .text import Text, new_cell

__all__ = ["BarChart", "Point", "Size"]

TICK_WIDTH = 11
VERTICAL_SPACING = 5
BAR_SPACING = 2
PRECISION = 4
DEFAULT_GRID_COLOR = Color.from_hex("272b40")


class _BarChartProperty[T]:
    def __set_name__(self, _, name):
        self.name = f"_{name}"

    def __get__(self, instance, _) -> T:
        if instance is None:
            return self  # type: ignore

        return getattr(instance, self.name)

    def __set__(self, instance, value):
        setattr(instance, self.name, value)
        instance._build_chart()


class BarChart(Gadget):
    r"""
    A bar chart gadget.

    Parameters
    ----------
    data : dict[str, float]
        Bar chart data.
    min_y : float | None, default: 0
        Minimum y-value of chart. If `None`, min_y will be minimum of all chart values.
    max_y : float | None, default: None
        Maximum y-value of chart. If `None`, max_y will be maximum of all chart values.
    bar_colors : list[Color] | None, default: None
        Color of each bar. If `None`, a rainbow gradient is used.
    chart_fg_color: Color, default: NEPTUNE_PRIMARY_FG
        Foreground color of chart.
    chart_bg_color: Color, default: NEPTUNE_PRIMARY_BG
        Background color of chart.
    alpha: float, default: 1.0
        Transparency of gadget.
    y_label : str | None, default: None
        Optional label for y-axis.
    show_grid_lines : bool, default: True
        Whether to show grid lines.
    grid_line_color : Color, default: DEFAULT_GRID_COLOR
        Color of grid lines if shown.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    data : data[str, float]
        Bar chart data.
    min_y : float | None
        Minimum y-value of chart. If `None`, min_y will be minimum of all chart values.
    max_y : float | None
        Maximum y-value of chart. If `None`, max_y will be maximum of all chart values.
    bar_colors : list[Color] | None
        Color of each bar. If `None`, a rainbow gradient is used.
    chart_fg_color: Color
        Foreground color of chart.
    chart_bg_color: Color
        Background color of chart.
    alpha: float
        Transparency of gadget.
    y_label : str | None
        Optional label for y-axis.
    show_grid_lines : bool
        Whether to show grid lines.
    grid_line_color : Color
        Color of grid lines if shown.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    data: _BarChartProperty[dict[str, float]] = _BarChartProperty()
    """Data for bar chart."""
    min_y: _BarChartProperty[float | None] = _BarChartProperty()
    """Minimum y-value of chart. If `None`, min_y will be minimum of chart values."""
    max_y: _BarChartProperty[float | None] = _BarChartProperty()
    """Maximum y-value of chart. If `None`, max_y will be maximum of chart values."""
    bar_colors: _BarChartProperty[list[Color] | None] = _BarChartProperty()
    """Color of each bar. If `None`, a rainbow gradient is used."""
    chart_fg_color: _BarChartProperty[Color] = _BarChartProperty()
    """Foreground color of bar chart."""
    chart_bg_color: _BarChartProperty[Color] = _BarChartProperty()
    """Background color of bar chart."""
    alpha: _BarChartProperty[float] = _BarChartProperty()
    """Transparency of gadget."""
    y_label: _BarChartProperty[str] = _BarChartProperty()
    """Optional label for y-axis."""
    show_grid_lines: _BarChartProperty[bool] = _BarChartProperty()
    """Whether to show grid lines."""
    grid_line_color: _BarChartProperty[Color] = _BarChartProperty()
    """Color of grid lines if shown."""

    def __init__(
        self,
        data: dict[str, float],
        *,
        min_y: float | None = 0,
        max_y: float | None = None,
        bar_colors: list[Color] | None = None,
        chart_fg_color: Color = NEPTUNE_PRIMARY_FG,
        chart_bg_color: Color = NEPTUNE_PRIMARY_BG,
        alpha: float = 1.0,
        y_label: str | None = None,
        show_grid_lines: bool = True,
        grid_line_color: Color = DEFAULT_GRID_COLOR,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._bars = Text(is_transparent=is_transparent)
        self._scroll_view = ScrollView(
            show_horizontal_bar=False,
            show_vertical_bar=False,
            allow_vertical_scroll=False,
            alpha=0,
            is_transparent=is_transparent,
        )
        self._scroll_view.view = self._bars
        self._y_ticks = Text(is_transparent=is_transparent)
        self._y_label_gadget = Text(is_transparent=is_transparent)
        self._container = Pane(
            size_hint={"height_hint": 1.0, "width_hint": 1.0},
            is_transparent=is_transparent,
        )
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self._container.add_gadgets(
            self._scroll_view, self._y_ticks, self._y_label_gadget
        )
        self.add_gadget(self._container)

        self.data = data
        self.min_y = min_y
        self.max_y = max_y
        self.bar_colors = bar_colors
        self.chart_fg_color = chart_fg_color
        self.chart_bg_color = chart_bg_color
        self.alpha = alpha
        self.y_label = y_label
        self.show_grid_lines = show_grid_lines
        self.grid_line_color = grid_line_color

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._container.is_transparent = self.is_transparent
        self._y_ticks.is_transparent = self.is_transparent
        self._y_label_gadget.is_transparent = self.is_transparent
        self._scroll_view.is_transparent = self.is_transparent
        self._bars.is_transparent = self.is_transparent

    def on_add(self):
        """Build chart on add."""
        super().on_add()
        self._build_chart()

    def _build_chart(self):
        """Build bar chart and set canvas and color arrays."""
        if not self.root:
            return

        self._container.bg_color = self.chart_bg_color
        self._container.alpha = self.alpha

        h, w = self.size
        has_y_label = self._y_label_gadget.is_enabled = self.y_label is not None
        if has_y_label:
            self._y_label_gadget.size = wcswidth(self.y_label), 1
            self._y_label_gadget.top = h // 2 - self._y_label_gadget.height // 2
            add_text(
                self._y_label_gadget.canvas[:, 0],
                self.y_label,
                fg_color=self.chart_fg_color,
                bg_color=self.chart_bg_color,
            )

        sv_left = has_y_label + TICK_WIDTH
        sv_width = w - sv_left
        self._scroll_view.pos = 0, sv_left
        self._scroll_view.size = h, sv_width

        nbars = len(self.data)
        min_bar_width = max(map(wcswidth, self.data))
        bars_width = max(
            BAR_SPACING + (min_bar_width + BAR_SPACING) * nbars,
            sv_width,
        )
        bar_width = (bars_width - BAR_SPACING * (nbars + 1)) // nbars
        self._bars.size = h, bars_width
        self._bars.canvas[:] = new_cell(
            fg_color=self.chart_fg_color, bg_color=self.chart_bg_color
        )

        min_y = min(self.data.values()) if self.min_y is None else self.min_y
        max_y = max(self.data.values()) if self.max_y is None else self.max_y

        chars = self._bars.chars[::-1]
        fg_colors = self._bars.canvas["fg_color"][::-1]

        # Regenerate Ticks
        self._y_ticks.left = has_y_label
        self._y_ticks.size = h, TICK_WIDTH
        self._y_ticks.canvas["fg_color"] = self.chart_fg_color
        self._y_ticks.canvas["bg_color"] = self.chart_bg_color
        self._y_ticks.chars[0, -1] = "┐"
        self._y_ticks.chars[1:-2, -1] = "│"
        self._y_ticks.chars[-2, -1] = "└"

        last_y = h - 3
        for row in range(last_y, -1, -VERTICAL_SPACING):
            y_label = lerp(max_y, min_y, row / last_y)
            self._y_ticks.add_str(
                f"{y_label:>{TICK_WIDTH - 2}.{PRECISION}g} ┤"[:TICK_WIDTH],
                pos=(row, 0),
            )
            if self.show_grid_lines:
                chars[row, :-1] = "─"
                fg_colors[row] = self.grid_line_color

        bar_colors = self.bar_colors or rainbow_gradient(nbars)
        y_delta = max_y - min_y

        for i, (label, value) in enumerate(self.data.items()):
            x1 = BAR_SPACING + (bar_width + BAR_SPACING) * i
            x2 = x1 + bar_width
            self._bars.add_str(label.center(bar_width), pos=(h - 1, x1))
            smooth_bar = smooth_vertical_bar(h - 3, (value - min_y) / y_delta, 0.5)
            chars.T[x1:x2, 2 : 2 + len(smooth_bar)] = smooth_bar
            fg_colors[2 : 2 + len(smooth_bar), x1:x2] = bar_colors[i]
            self._bars.canvas["style"][::-1].T[x1:x2, 2] = Style.REVERSE
        chars[1, :-1] = "─"
        chars[1, -1] = "┐"

    def on_size(self):
        """Rebuild bar chart."""
        self._build_chart()



"""Color data structures."""

import re
from dataclasses import dataclass
from typing import Final, Literal, NamedTuple, Self

from .._style import Style

__all__ = ["AColor", "Color", "ColorTheme", "SyntaxHighlightTheme"]

_HEXCODE_RE: Final = re.compile(r"^#?[0-9a-fA-F]{6}$")
_AHEXCODE_RE: Final = re.compile(r"^#?[0-9a-fA-F]{8}$")

type Hexcode = str
"""
A hexcode string for 24-bit color. Six hex digits optionally preceded by "#", e.g.,
"#abc123".
"""

type AHexcode = str
"""
A hexcode string for a 32-bit color. Six or eight hex digits optionally preceded by
"#", e.g., "#abcd1234". If eight hex digits are given, the last two digits specify the
alpha channel, otherwise, the alpha channel value defaults to 255 so that the six digit
string ``"#abc123`` is equivalent to ``"#abc123ff"``.
"""


def validate_hexcode(hexcode: Hexcode) -> bool:
    return bool(_HEXCODE_RE.match(hexcode))


def validate_ahexcode(ahexcode: AHexcode) -> bool:
    return validate_hexcode(ahexcode) or bool(_AHEXCODE_RE.match(ahexcode))


def _to_hex(channel: int):
    return f"{channel:02x}"


class Color(NamedTuple):
    """
    A 24-bit color.

    Parameters
    ----------
    red : int
        The red component.
    green : int
        The green component.
    blue : int
        The blue component.

    Attributes
    ----------
    red : int
        The red component.
    green : int
        The green component.
    blue : int
        The blue component.

    Methods
    -------
    from_hex(hexcode)
        Create a :class:`Color` from a hex code.
    to_hex()
        Return color's hexcode.
    """

    red: int
    green: int
    blue: int

    @classmethod
    def from_hex(cls, hexcode: Hexcode) -> Self:
        """
        Create a :class:`Color` from a hex code.

        Parameters
        ----------
        hexcode : Hexcode
            A color hex code.

        Returns
        -------
        Color
            A new color.
        """
        if not validate_hexcode(hexcode):
            raise ValueError(f"{hexcode!r} is not a valid hex code.")

        digits = hexcode.removeprefix("#")
        return cls(int(digits[:2], 16), int(digits[2:4], 16), int(digits[4:], 16))

    def to_hex(self) -> Hexcode:
        """
        Return color's hexcode.

        Returns
        -------
        Hexcode
            The hexcode of the color.
        """
        hexcode = "".join(_to_hex(channel) for channel in self)
        return f"#{hexcode}"


class AColor(NamedTuple):
    """
    A 32-bit color with an alpha channel.

    Parameters
    ----------
    red : int
        The red component.
    green : int
        The green component.
    blue : int
        The blue component.
    alpha : int, default: 255
        The alpha component.

    Attributes
    ----------
    red : int
        The red component.
    green : int
        The green component.
    blue : int
        The blue component.
    alpha : int
        The alpha component.

    Methods
    -------
    from_hex(hexcode)
        Create an :class:`AColor` from a hex code.
    to_hex()
        Return color's hexcode.
    """

    red: int
    green: int
    blue: int
    alpha: int = 255

    @classmethod
    def from_hex(cls, hexcode: AHexcode) -> Self:
        """
        Create an :class:`AColor` from a hex code.

        If the hexcode only has six hex digits, the alpha channel's value will default
        to 255.

        Parameters
        ----------
        hexcode : AHexcode
            A color hex code.

        Returns
        -------
        AColor
            A new color with an alpha channel.
        """
        if not validate_ahexcode(hexcode):
            raise ValueError(f"{hexcode!r} is not a valid hex code")

        digits = hexcode.removeprefix("#")
        return cls(
            int(digits[:2], 16),
            int(digits[2:4], 16),
            int(digits[4:6], 16),
            int(digits[6:] or "ff", 16),
        )

    def to_hex(self) -> AHexcode:
        """
        Return color's hexcode.

        Returns
        -------
        AHexcode
            The hexcode of the color.
        """
        hexcode = "".join(_to_hex(channel) for channel in self)
        return f"#{hexcode}"


type ColorThemeColor = Literal[
    "primary_fg",
    "primary_bg",
    "text_pad_line_highlight_fg",
    "text_pad_line_highlight_bg",
    "text_pad_selection_highlight_fg",
    "text_pad_selection_highlight_bg",
    "textbox_primary_fg",
    "textbox_primary_bg",
    "textbox_selection_highlight_fg",
    "textbox_selection_highlight_bg",
    "textbox_placeholder_fg",
    "textbox_placeholder_bg",
    "button_normal_fg",
    "button_normal_bg",
    "button_hover_fg",
    "button_hover_bg",
    "button_press_fg",
    "button_press_bg",
    "button_disallowed_fg",
    "button_disallowed_bg",
    "menu_item_hover_fg",
    "menu_item_hover_bg",
    "menu_item_selected_fg",
    "menu_item_selected_bg",
    "menu_item_disallowed_fg",
    "menu_item_disallowed_bg",
    "titlebar_normal_fg",
    "titlebar_normal_bg",
    "titlebar_inactive_fg",
    "titlebar_inactive_bg",
    "data_table_sort_indicator_fg",
    "data_table_sort_indicator_bg",
    "data_table_hover_fg",
    "data_table_hover_bg",
    "data_table_stripe_fg",
    "data_table_stripe_bg",
    "data_table_stripe_hover_fg",
    "data_table_stripe_hover_bg",
    "data_table_selected_fg",
    "data_table_selected_bg",
    "data_table_selected_hover_fg",
    "data_table_selected_hover_bg",
    "progress_bar_fg",
    "progress_bar_bg",
    "markdown_link_fg",
    "markdown_link_bg",
    "markdown_link_hover_fg",
    "markdown_link_hover_bg",
    "markdown_inline_code_fg",
    "markdown_inline_code_bg",
    "markdown_quote_fg",
    "markdown_quote_bg",
    "markdown_title_fg",
    "markdown_title_bg",
    "markdown_image_fg",
    "markdown_image_bg",
    "markdown_block_code_bg",
    "markdown_quote_block_code_bg",
    "markdown_header_bg",
    "scroll_view_scrollbar",
    "scroll_view_indicator_normal",
    "scroll_view_indicator_hover",
    "scroll_view_indicator_press",
    "window_border_normal",
    "window_border_inactive",
]

type ColorTheme = dict[ColorThemeColor, Hexcode]
"""
Colors for themable gadgets.

Missing colors will use the default color theme.
"""


@dataclass
class SyntaxHighlightTheme:
    """Syntax highlight theme."""

    default_fg: Color
    """Default foreground color."""
    default_bg: Color
    """Default background color."""
    cursor_fg: Color
    """Default cursor foreground color."""
    cursor_bg: Color
    """Default cursor background color."""
    active_line: Color
    """The background color of the cursor's current line."""
    selection: Color
    """The background color of selected text."""
    highlights: dict[str, tuple[Style, Color | None]]
    """A dictionary of tree sitter query names to (style, Color) tuples."""

    def __init__(
        self,
        default_fg: Hexcode,
        default_bg: Hexcode,
        cursor_fg: Hexcode,
        cursor_bg: Hexcode,
        active_line: Hexcode,
        selection: Hexcode,
        highlights: dict[str, tuple[Style, Hexcode | None]],
    ):
        self.default_fg = Color.from_hex(default_fg)
        self.default_bg = Color.from_hex(default_bg)
        self.cursor_fg = Color.from_hex(cursor_fg)
        self.cursor_bg = Color.from_hex(cursor_bg)
        self.active_line = Color.from_hex(active_line)
        self.selection = Color.from_hex(selection)
        self.highlights = {
            k: (style, None if hexcode is None else Color.from_hex(hexcode))
            for k, (style, hexcode) in highlights.items()
        }



from .terminal._fbuf cimport fbuf

cdef:
    struct qnode:
        unsigned char[3] srgb
        unsigned long pop
        unsigned int qlink
        unsigned int cidx

    struct onode:
        (qnode *)[8] q

    struct qstate:
        qnode *qnodes
        onode *onodes
        unsigned int dynnodes_free
        unsigned int dynnodes_total
        unsigned onodes_free
        unsigned onodes_total
        unsigned long ncolors
        unsigned char *table

    int sixel(
        fbuf *f,
        qstate *qs,
        unsigned char[:, :, ::1] texture,
        unsigned char[:, :, ::1] stexture,
        unsigned int aspect_h,
        unsigned int aspect_w,
        size_t oy,
        size_t ox,
        size_t h,
        size_t w,
    )

    class OctTree:
        cdef qstate qs



"""A toggle button gadget."""

from collections.abc import Callable, Hashable

from .behaviors.themable import Themable
from .behaviors.toggle_button_behavior import (
    ButtonState,
    ToggleButtonBehavior,
    ToggleState,
)
from .gadget import Gadget, Point, Pointlike, PosHint, Size, SizeHint, Sizelike
from .pane import Pane
from .text import Text

__all__ = ["ButtonState", "Point", "Size", "ToggleButton", "ToggleState"]

CHECK_OFF = "☐ "
CHECK_ON = "🗹 "
TOGGLE_OFF = "○ "
TOGGLE_ON = "● "


class ToggleButton(Themable, ToggleButtonBehavior, Gadget):
    r"""
    A toggle button gadget.

    Without a group, a toggle button behaves like a checkbox (either on or off). With a
    group, it behaves like a radio button (only a single button in a group is allowed to
    be in the "on" state).

    Parameters
    ----------
    label : str, default: ""
        Toggle button label.
    callback : Callable[[ToggleState], None] | None, default: None
        Called when toggle state changes. The new state is provided as first argument.
    alpha : float, default: 1.0
        Transparency of gadget.
    group : Hashable | None, default: None
        If a group is provided, only one button in a group can be in the on state.
    allow_no_selection : bool, default: False
        If a group is provided, setting this to true allows no selection, i.e.,
        every button can be in the off state.
    always_release : bool, default: False
        Whether a mouse up event outside the button will trigger it.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    label : str
        Toggle button label.
    callback : Callable[[ToggleState], None] | None
        Called when toggle state changes.
    alpha : float
        Transparency of gadget.
    group : Hashable | None
        If a group is provided, only one button in a group can be in the on state.
    allow_no_selection : bool
        If true and button is in a group, every button can be in the off state.
    toggle_state : ToggleState
        Toggle state of button.
    always_release : bool
        Whether a mouse up event outside the button will trigger it.
    button_state : ButtonState
        Current button state.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    get_color(color_name)
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    on_toggle()
        Triggled on toggle state change.
    update_off()
        Paint the off state.
    update_on()
        Paint the on state.
    on_release()
        Triggered when a button is released.
    update_normal()
        Paint the normal state.
    update_hover()
        Paint the hover state.
    update_down()
        Paint the down state.
    update_disallowed()
        Paint the disallowed state.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        label: str = "",
        callback: Callable[[ToggleState], None] | None = None,
        alpha: float = 1.0,
        group: None | Hashable = None,
        allow_no_selection: bool = False,
        always_release: bool = False,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._pane = Pane(
            size_hint={"height_hint": 1.0, "width_hint": 1.0},
            is_transparent=is_transparent,
        )
        self._label = Text(
            pos_hint={"y_hint": 0.5, "anchor": "left"}, is_transparent=True
        )
        self._label_text = label
        super().__init__(
            group=group,
            allow_no_selection=allow_no_selection,
            always_release=always_release,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.add_gadgets(self._pane, self._label)
        self.callback = callback
        self.alpha = alpha

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._pane.is_transparent = self.is_transparent

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._pane.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._pane.alpha = alpha

    @property
    def label(self) -> str:
        """Toggle button label."""
        return self._label_text

    @label.setter
    def label(self, label: str):
        self._label_text = label

        if self.group is None:
            on, off = CHECK_ON, CHECK_OFF
        else:
            on, off = TOGGLE_ON, TOGGLE_OFF

        if self.toggle_state == "on":
            prefix = on
        else:
            prefix = off

        self._label.set_text(prefix + label)

    def update_theme(self):
        """Paint the gadget with current theme."""
        getattr(self, f"update_{self.button_state}")()

    def update_on(self):
        """Paint the on state."""
        if self.group is None:
            prefix = CHECK_ON
        else:
            prefix = TOGGLE_ON
        self._label.set_text(prefix + self._label_text)

    def update_off(self):
        """Paint the off state."""
        if self.group is None:
            prefix = CHECK_OFF
        else:
            prefix = TOGGLE_OFF
        self._label.set_text(prefix + self._label_text)

    def update_normal(self):
        """Paint the normal state."""
        self._pane.bg_color = self.get_color("button_normal_bg")
        self._label.canvas["fg_color"] = self.get_color("button_normal_fg")

    def update_hover(self):
        """Paint the hover state."""
        self._pane.bg_color = self.get_color("button_hover_bg")
        self._label.canvas["fg_color"] = self.get_color("button_hover_fg")

    def update_down(self):
        """Paint the down state."""
        self._pane.bg_color = self.get_color("button_press_bg")
        self._label.canvas["fg_color"] = self.get_color("button_press_fg")

    def update_disallowed(self):
        """Paint the disallowed state."""
        self._pane.bg_color = self.get_color("button_disallowed_bg")
        self._label.canvas["fg_color"] = self.get_color("button_disallowed_fg")

    def on_toggle(self):
        """Call callback on toggle state change."""
        if self.root is None:
            return

        self.label = self.label  # Update radio button/checkbox
        if self.callback is not None:
            self.callback(self.toggle_state)



"""Tools for graphics."""

from pathlib import Path
from typing import Literal, cast

import cv2
import numpy as np

from .array_types import RGBA_2D
from .geometry import Point, Region, Sizelike, rect_slice

__all__ = [
    "Interpolation",
    "composite",
    "read_buffer",
    "read_texture",
    "resize_texture",
]

Interpolation = Literal["nearest", "linear", "cubic", "area", "lanczos"]
"""Interpolation methods for resizing graphic gadgets."""

_INTERPOLATION_TO_CV_ENUM = {
    "linear": cv2.INTER_LINEAR,
    "cubic": cv2.INTER_CUBIC,
    "area": cv2.INTER_AREA,
    "lanczos": cv2.INTER_LANCZOS4,
    "nearest": cv2.INTER_NEAREST,
}


def read_buffer(buffer: bytes) -> RGBA_2D:
    """
    Return a uint8 RGBA numpy array from bytes.

    Parameters
    ----------
    path : Path
        Path to image.

    Returns
    -------
    RGBA_2D
        An uint8 RGBA array of the image.
    """
    decoded = np.frombuffer(buffer, dtype=np.uint8)
    image = cv2.imdecode(decoded, cv2.IMREAD_UNCHANGED)
    image = cast(np.ndarray, image)
    return _texture_to_RGBA_2D(image)


def read_texture(path: Path) -> RGBA_2D:
    """
    Return a uint8 RGBA numpy array from a path to an image.

    Parameters
    ----------
    path : Path
        Path to image.

    Returns
    -------
    RGBA_2D
        An uint8 RGBA array of the image.
    """
    image = cv2.imread(str(path.absolute()), cv2.IMREAD_UNCHANGED)
    image = cast(np.ndarray, image)
    return _texture_to_RGBA_2D(image)


def _texture_to_RGBA_2D(texture: np.ndarray) -> RGBA_2D:
    if texture.dtype == np.dtype(np.uint16):
        texture = (texture // 257).astype(np.uint8)
    elif texture.dtype == np.dtype(np.float32):
        texture = (texture * 255).astype(np.uint8)

    if texture.ndim == 2:
        texture = np.dstack((texture, texture, texture))

    # Add an alpha channel if there isn't one.
    h, w, c = texture.shape
    if c == 3:
        default_alpha_channel = np.full((h, w, 1), 255, dtype=np.uint8)
        texture = np.dstack((texture, default_alpha_channel))

    return cv2.cvtColor(texture, cv2.COLOR_BGRA2RGBA)  # type: ignore


def resize_texture(
    texture: RGBA_2D,
    size: Sizelike,
    interpolation: Interpolation = "linear",
    out: RGBA_2D | None = None,
) -> RGBA_2D:
    """
    Resize texture.

    Parameters
    ----------
    texture : RGBA_2D
        An RGBA texture to resize.
    size : Sizelike
        The new size of the texture.
    interpolation : Interpolation, default: "linear"
        Interpolation used when resizing texture.
    out : RGBA_2D | None, default: None
        Optional output array. If None, a new array is created.

    Returns
    -------
    RGBA_2D
        A new uint8 RGBA array.
    """
    old_h, old_w = texture.shape[:2]
    h, w = size
    if old_h == 0 or old_w == 0 or h == 0 or w == 0:
        return np.zeros((h, w, 4), np.uint8)  # type: ignore
    return cv2.resize(
        texture, (w, h), dst=out, interpolation=_INTERPOLATION_TO_CV_ENUM[interpolation]
    )  # type: ignore


def composite(
    source: RGBA_2D,
    dest: RGBA_2D,
    pos: Point = Point(0, 0),
    mask_mode: bool = False,
) -> None:
    """
    Composite source texture onto destination texture at given position.

    If `mask_mode` is true, source alpha values less than 255 are ignored.

    Parameters
    ----------
    source : RGBA_2D
        The texture to composite.
    dest : RGBA_2D
        The texture on which the source is painted.
    pos : Pointlike, default: Point(0, 0)
        Position of the source on the destination.
    mask_mode : bool, default: False
        Whether to ignore alpha values less than 255.
    """
    sh, sw, _ = source.shape
    dh, dw, _ = dest.shape

    dest_reg = Region.from_rect((0, 0), (dh, dw))
    source_reg = Region.from_rect(pos, (sh, sw))

    if intersection := source_reg & dest_reg:
        rpos, size = next(intersection.rects())
        dest_tex = dest[rect_slice(rpos, size)]
        source_tex = source[rect_slice(rpos - pos, size)]
        source_alpha = source_tex[..., 3]

        if mask_mode:
            mask = source_alpha == 255
            dest_tex[mask] = source_tex[mask]
        else:
            buffer = np.subtract(source_tex, dest_tex, dtype=float)
            buffer *= source_alpha[..., None]
            buffer /= 255
            np.add(buffer, dest_tex, out=dest_tex, casting="unsafe")



"""Mapping from VT100 (ANSI) escape sequences to the corresponding keys."""

from .events import Key

ANSI_ESCAPES: dict[bytes, tuple[Key, bool, bool, bool]] = {
    # ansi: (key, alt, ctrl, shift)
    b"\x00": (" ", False, True, False),
    b"\x01": ("a", False, True, False),
    b"\x02": ("b", False, True, False),
    b"\x03": ("c", False, True, False),
    b"\x04": ("d", False, True, False),
    b"\x05": ("e", False, True, False),
    b"\x06": ("f", False, True, False),
    b"\x07": ("g", False, True, False),
    b"\x08": ("h", False, True, False),
    b"\x09": ("tab", False, False, False),  # ctrl + i
    b"\x0a": ("j", False, True, False),
    b"\x0b": ("k", False, True, False),
    b"\x0c": ("l", False, True, False),
    b"\x0d": ("enter", False, False, False),  # ctrl + m
    b"\x0e": ("n", False, True, False),
    b"\x0f": ("o", False, True, False),
    b"\x10": ("p", False, True, False),
    b"\x11": ("q", False, True, False),
    b"\x12": ("r", False, True, False),
    b"\x13": ("s", False, True, False),
    b"\x14": ("t", False, True, False),
    b"\x15": ("u", False, True, False),
    b"\x16": ("v", False, True, False),
    b"\x17": ("w", False, True, False),
    b"\x18": ("x", False, True, False),
    b"\x19": ("y", False, True, False),
    b"\x1a": ("z", False, True, False),
    b"\x1b": ("escape", False, False, False),
    b"\x1c": ("\\", False, True, False),
    b"\x1d": ("]", False, True, False),
    b"\x1e": ("^", False, True, False),
    b"\x1f": ("-", False, True, False),
    b"\x7f": ("backspace", False, False, False),  # not ctrl + h on WSL
    b"\x9b": ("escape", False, False, True),
    b"\x1b\x00": (" ", True, True, False),
    b"\x1b\x01": ("a", True, True, False),
    b"\x1b\x02": ("b", True, True, False),
    b"\x1b\x03": ("c", True, True, False),
    b"\x1b\x04": ("d", True, True, False),
    b"\x1b\x05": ("e", True, True, False),
    b"\x1b\x06": ("f", True, True, False),
    b"\x1b\x07": ("g", True, True, False),
    b"\x1b\x08": ("h", True, True, False),
    b"\x1b\x09": ("i", True, True, False),
    b"\x1b\x0a": ("j", True, True, False),
    b"\x1b\x0b": ("k", True, True, False),
    b"\x1b\x0c": ("l", True, True, False),
    b"\x1b\x0d": ("m", True, True, False),
    b"\x1b\x0e": ("n", True, True, False),
    b"\x1b\x0f": ("o", True, True, False),
    b"\x1b\x10": ("p", True, True, False),
    b"\x1b\x11": ("q", True, True, False),
    b"\x1b\x12": ("r", True, True, False),
    b"\x1b\x13": ("s", True, True, False),
    b"\x1b\x14": ("t", True, True, False),
    b"\x1b\x15": ("u", True, True, False),
    b"\x1b\x16": ("v", True, True, False),
    b"\x1b\x17": ("w", True, True, False),
    b"\x1b\x18": ("x", True, True, False),
    b"\x1b\x19": ("y", True, True, False),
    b"\x1b\x1a": ("z", True, True, False),
    b"\x1b\x1b": ("#", True, True, False),
    b"\x1b\x1c": ("$", True, True, False),
    b"\x1b\x1d": ("%", True, True, False),
    b"\x1b\x1e": ("^", True, True, False),
    b"\x1b\x1f": ("&", True, True, False),
    b"\x1b\x7f": ("*", True, True, False),
    b"\x1bOA": ("up", False, False, False),
    b"\x1bOB": ("down", False, False, False),
    b"\x1bOC": ("right", False, False, False),
    b"\x1bOD": ("left", False, False, False),
    b"\x1bOF": ("end", False, False, False),
    b"\x1bOH": ("home", False, False, False),
    b"\x1bOP": ("f1", False, False, False),
    b"\x1bOQ": ("f2", False, False, False),
    b"\x1bOR": ("f3", False, False, False),
    b"\x1bOS": ("f4", False, False, False),
    b"\x1bOc": ("right", False, True, False),
    b"\x1bOd": ("left", False, True, False),
    b"\x1b[[A": ("f1", False, False, False),
    b"\x1b[[B": ("f2", False, False, False),
    b"\x1b[[C": ("f3", False, False, False),
    b"\x1b[[D": ("f4", False, False, False),
    b"\x1b[[E": ("f5", False, False, False),
    b"\x1b[A": ("up", False, False, False),
    b"\x1b[B": ("down", False, False, False),
    b"\x1b[C": ("right", False, False, False),
    b"\x1b[D": ("left", False, False, False),
    b"\x1b[F": ("end", False, False, False),
    b"\x1b[H": ("home", False, False, False),
    b"\x1b[Z": ("tab", False, False, True),
    b"\x1b[5A": ("up", False, True, False),
    b"\x1b[5B": ("down", False, True, False),
    b"\x1b[5C": ("right", False, True, False),
    b"\x1b[5D": ("left", False, True, False),
    b"\x1b[1;6P": ("f13", False, True, False),
    b"\x1b[1;6Q": ("f14", False, True, False),
    b"\x1b[1;6R": ("f15", False, True, False),
    b"\x1b[1;6S": ("f16", False, True, False),
    b"\x1b[1~": ("home", False, False, False),
    b"\x1b[2~": ("insert", False, False, False),
    b"\x1b[3~": ("delete", False, False, False),
    b"\x1b[4~": ("end", False, False, False),
    b"\x1b[5~": ("page_up", False, False, False),
    b"\x1b[6~": ("page_down", False, False, False),
    b"\x1b[7~": ("home", False, False, False),
    b"\x1b[8~": ("end", False, False, False),
    b"\x1b[~": ("tab", False, False, True),
    b"\x1b[11~": ("f1", False, False, False),
    b"\x1b[12~": ("f2", False, False, False),
    b"\x1b[13~": ("f3", False, False, False),
    b"\x1b[14~": ("f4", False, False, False),
    b"\x1b[15~": ("f5", False, False, False),
    b"\x1b[17~": ("f6", False, False, False),
    b"\x1b[18~": ("f7", False, False, False),
    b"\x1b[19~": ("f8", False, False, False),
    b"\x1b[20~": ("f9", False, False, False),
    b"\x1b[21~": ("f10", False, False, False),
    b"\x1b[23~": ("f11", False, False, False),
    b"\x1b[24~": ("f12", False, False, False),
    b"\x1b[25~": ("f13", False, False, False),
    b"\x1b[26~": ("f14", False, False, False),
    b"\x1b[28~": ("f15", False, False, False),
    b"\x1b[29~": ("f16", False, False, False),
    b"\x1b[31~": ("f17", False, False, False),
    b"\x1b[32~": ("f18", False, False, False),
    b"\x1b[33~": ("f19", False, False, False),
    b"\x1b[34~": ("f20", False, False, False),
    b"\x1b[15;2~": ("f17", False, False, False),
    b"\x1b[17;2~": ("f18", False, False, False),
    b"\x1b[18;2~": ("f19", False, False, False),
    b"\x1b[19;2~": ("f20", False, False, False),
    b"\x1b[20;2~": ("f21", False, False, False),
    b"\x1b[21;2~": ("f22", False, False, False),
    b"\x1b[23;2~": ("f23", False, False, False),
    b"\x1b[24;2~": ("f24", False, False, False),
    b"\x1b[62~": ("scroll_up", False, False, False),
    b"\x1b[63~": ("scroll_down", False, False, False),
    b"\x1b[3;2~": ("delete", False, False, True),
    b"\x1b[5;2~": ("page_up", False, False, True),
    b"\x1b[6;2~": ("page_down", False, False, True),
    b"\x1b[2;3~": ("insert", True, False, False),
    b"\x1b[3;3~": ("delete", True, False, False),
    b"\x1b[5;3~": ("page_up", True, False, False),
    b"\x1b[6;3~": ("page_down", True, False, False),
    b"\x1b[2;4~": ("insert", True, False, True),
    b"\x1b[3;4~": ("delete", True, False, True),
    b"\x1b[5;4~": ("page_up", True, False, True),
    b"\x1b[6;4~": ("page_down", True, False, True),
    b"\x1b[3;5~": ("delete", False, True, False),
    b"\x1b[5;5~": ("page_up", False, True, False),
    b"\x1b[6;5~": ("page_down", False, True, False),
    b"\x1b[15;5~": ("f5", False, True, False),
    b"\x1b[17;5~": ("f6", False, True, False),
    b"\x1b[18;5~": ("f7", False, True, False),
    b"\x1b[19;5~": ("f8", False, True, False),
    b"\x1b[20;5~": ("f9", False, True, False),
    b"\x1b[21;5~": ("f10", False, True, False),
    b"\x1b[23;5~": ("f11", False, True, False),
    b"\x1b[24;5~": ("f12", False, True, False),
    b"\x1b[3;6~": ("delete", False, True, True),
    b"\x1b[5;6~": ("page_up", False, True, True),
    b"\x1b[6;6~": ("page_down", False, True, True),
    b"\x1b[15;6~": ("f17", False, True, False),
    b"\x1b[17;6~": ("f18", False, True, False),
    b"\x1b[18;6~": ("f19", False, True, False),
    b"\x1b[19;6~": ("f20", False, True, False),
    b"\x1b[20;6~": ("f21", False, True, False),
    b"\x1b[21;6~": ("f22", False, True, False),
    b"\x1b[23;6~": ("f23", False, True, False),
    b"\x1b[24;6~": ("f24", False, True, False),
    b"\x1b[2;7~": ("insert", True, True, False),
    b"\x1b[5;7~": ("page_up", True, True, False),
    b"\x1b[6;7~": ("page_down", True, True, False),
    b"\x1b[2;8~": ("insert", True, True, True),
    b"\x1b[5;8~": ("page_up", True, True, True),
    b"\x1b[6;8~": ("page_down", True, True, True),
    b"\x1b[1;2A": ("up", False, False, True),
    b"\x1b[1;2B": ("down", False, False, True),
    b"\x1b[1;2C": ("right", False, False, True),
    b"\x1b[1;2D": ("left", False, False, True),
    b"\x1b[1;2F": ("end", False, False, True),
    b"\x1b[1;2H": ("home", False, False, True),
    b"\x1b[1;2P": ("f13", False, False, False),
    b"\x1b[1;2Q": ("f14", False, False, False),
    b"\x1b[1;2R": ("f15", False, False, False),
    b"\x1b[1;2S": ("f16", False, False, False),
    b"\x1b[1;3A": ("up", True, False, False),
    b"\x1b[1;3B": ("down", True, False, False),
    b"\x1b[1;3C": ("right", True, False, False),
    b"\x1b[1;3D": ("left", True, False, False),
    b"\x1b[1;3F": ("end", True, False, False),
    b"\x1b[1;3H": ("home", True, False, False),
    b"\x1b[1;4A": ("up", True, False, True),
    b"\x1b[1;4B": ("down", True, False, True),
    b"\x1b[1;4C": ("right", True, False, True),
    b"\x1b[1;4D": ("left", True, False, True),
    b"\x1b[1;4F": ("end", True, False, True),
    b"\x1b[1;4H": ("home", True, False, True),
    b"\x1b[1;5A": ("up", False, True, False),
    b"\x1b[1;5B": ("down", False, True, False),
    b"\x1b[1;5C": ("right", False, True, False),
    b"\x1b[1;5D": ("left", False, True, False),
    b"\x1b[1;5F": ("end", False, True, False),
    b"\x1b[1;5H": ("home", False, True, False),
    b"\x1b[1;5P": ("f1", False, True, False),
    b"\x1b[1;5Q": ("f2", False, True, False),
    b"\x1b[1;5R": ("f3", False, True, False),
    b"\x1b[1;5S": ("f4", False, True, False),
    b"\x1b[1;6A": ("up", False, True, True),
    b"\x1b[1;6B": ("down", False, True, True),
    b"\x1b[1;6C": ("right", False, True, True),
    b"\x1b[1;6D": ("left", False, True, True),
    b"\x1b[1;6F": ("end", False, True, True),
    b"\x1b[1;6H": ("home", False, True, True),
    b"\x1b[1;7A": ("up", True, True, False),
    b"\x1b[1;7B": ("down", True, True, False),
    b"\x1b[1;7C": ("right", True, True, False),
    b"\x1b[1;7D": ("left", True, True, False),
    b"\x1b[1;7F": ("end", True, True, False),
    b"\x1b[1;7H": ("home", True, True, False),
    b"\x1b[1;8A": ("up", True, True, True),
    b"\x1b[1;8B": ("down", True, True, True),
    b"\x1b[1;8C": ("right", True, True, True),
    b"\x1b[1;8D": ("left", True, True, True),
    b"\x1b[1;8F": ("end", True, True, True),
    b"\x1b[1;8H": ("home", True, True, True),
    b"\x1b[1;9A": ("up", True, False, False),
    b"\x1b[1;9B": ("down", True, False, False),
    b"\x1b[1;9C": ("right", True, False, False),
    b"\x1b[1;9D": ("left", True, False, False),
    b"\x1b[1;5p": ("0", False, True, False),
    b"\x1b[1;5q": ("1", False, True, False),
    b"\x1b[1;5r": ("2", False, True, False),
    b"\x1b[1;5s": ("3", False, True, False),
    b"\x1b[1;5t": ("4", False, True, False),
    b"\x1b[1;5u": ("5", False, True, False),
    b"\x1b[1;5v": ("6", False, True, False),
    b"\x1b[1;5w": ("7", False, True, False),
    b"\x1b[1;5x": ("8", False, True, False),
    b"\x1b[1;5y": ("9", False, True, False),
    b"\x1b[1;6p": ("0", False, True, True),
    b"\x1b[1;6q": ("1", False, True, True),
    b"\x1b[1;6r": ("2", False, True, True),
    b"\x1b[1;6s": ("3", False, True, True),
    b"\x1b[1;6t": ("4", False, True, True),
    b"\x1b[1;6u": ("5", False, True, True),
    b"\x1b[1;6v": ("6", False, True, True),
    b"\x1b[1;6w": ("7", False, True, True),
    b"\x1b[1;6x": ("8", False, True, True),
    b"\x1b[1;6y": ("9", False, True, True),
    b"\x1b[1;7p": ("0", True, True, False),
    b"\x1b[1;7q": ("1", True, True, False),
    b"\x1b[1;7r": ("2", True, True, False),
    b"\x1b[1;7s": ("3", True, True, False),
    b"\x1b[1;7t": ("4", True, True, False),
    b"\x1b[1;7u": ("5", True, True, False),
    b"\x1b[1;7v": ("6", True, True, False),
    b"\x1b[1;7w": ("7", True, True, False),
    b"\x1b[1;7x": ("8", True, True, False),
    b"\x1b[1;7y": ("9", True, True, False),
    b"\x1b[1;8p": ("0", True, True, True),
    b"\x1b[1;8q": ("1", True, True, True),
    b"\x1b[1;8r": ("2", True, True, True),
    b"\x1b[1;8s": ("3", True, True, True),
    b"\x1b[1;8t": ("4", True, True, True),
    b"\x1b[1;8u": ("5", True, True, True),
    b"\x1b[1;8v": ("6", True, True, True),
    b"\x1b[1;8w": ("7", True, True, True),
    b"\x1b[1;8x": ("8", True, True, True),
    b"\x1b[1;8y": ("9", True, True, True),
    # WSL
    b"\x1b[15;3~": ("f5", True, False, False),
    b"\x1b[17;3~": ("f6", True, False, False),
    b"\x1b[18;3~": ("f7", True, False, False),
    b"\x1b[19;3~": ("f8", True, False, False),
    b"\x1b[20;3~": ("f9", True, False, False),
    b"\x1b[21;3~": ("f10", True, False, False),
    b"\x1b[23;3~": ("f11", True, False, False),
    b"\x1b[24;3~": ("f12", True, False, False),
    b"\x1b[15;4~": ("f5", True, False, True),
    b"\x1b[17;4~": ("f6", True, False, True),
    b"\x1b[18;4~": ("f7", True, False, True),
    b"\x1b[19;4~": ("f8", True, False, True),
    b"\x1b[20;4~": ("f9", True, False, True),
    b"\x1b[21;4~": ("f10", True, False, True),
    b"\x1b[23;4~": ("f11", True, False, True),
    b"\x1b[24;4~": ("f12", True, False, True),
    b"\x1b[15;7~": ("f5", True, True, False),
    b"\x1b[17;7~": ("f6", True, True, False),
    b"\x1b[18;7~": ("f7", True, True, False),
    b"\x1b[19;7~": ("f8", True, True, False),
    b"\x1b[20;7~": ("f9", True, True, False),
    b"\x1b[21;7~": ("f10", True, True, False),
    b"\x1b[23;7~": ("f11", True, True, False),
    b"\x1b[24;7~": ("f12", True, True, False),
    b"\x1b[15;8~": ("f5", True, True, True),
    b"\x1b[17;8~": ("f6", True, True, True),
    b"\x1b[18;8~": ("f7", True, True, True),
    b"\x1b[19;8~": ("f8", True, True, True),
    b"\x1b[20;8~": ("f9", True, True, True),
    b"\x1b[21;8~": ("f10", True, True, True),
    b"\x1b[23;8~": ("f11", True, True, True),
    b"\x1b[24;8~": ("f12", True, True, True),
    b"\x1b[1;3P": ("f1", True, False, False),
    b"\x1b[1;3Q": ("f2", True, False, False),
    b"\x1b[1;3R": ("f3", True, False, False),
    b"\x1b[1;4P": ("f1", True, False, True),
    b"\x1b[1;4Q": ("f2", True, False, True),
    b"\x1b[1;4R": ("f3", True, False, True),
    b"\x1b[1;4S": ("f4", True, False, True),
    b"\x1b[1;7P": ("f1", True, True, False),
    b"\x1b[1;7Q": ("f2", True, True, False),
    b"\x1b[1;7R": ("f3", True, True, False),
    b"\x1b[1;7S": ("f4", True, True, False),
    b"\x1b[1;8P": ("f1", True, True, True),
    b"\x1b[1;8Q": ("f2", True, True, True),
    b"\x1b[1;8R": ("f3", True, True, True),
    b"\x1b[1;8S": ("f4", True, True, True),
}
